//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include "LBFGSFitTrilinearModel.h"
#include <iostream> 
#include "Helpers.h" 
#include "ProjectionData.h"
#include <float.h>
#include <sstream>

LBFGSFitTrilinearModel::LBFGSFitTrilinearModel(void) : m_x(NULL),
	mobjProbabilityModel(NULL)
{
}

LBFGSFitTrilinearModel::~LBFGSFitTrilinearModel(void)
{
    if (m_x != NULL) {
        lbfgs_free(m_x);
        m_x = NULL;
    }
	if (mobjData != NULL)
		delete mobjData ; 
	mobjData = NULL ; 
	if (mobjProbabilityModel != NULL)
		delete mobjProbabilityModel ; 
	mobjProbabilityModel = NULL ; 
}

 int LBFGSFitTrilinearModel::FitModel(const std::string &mrcFile, 
	const std::string &analysisName, 
	const std::string &rotationFile,
	const std::string startModelMRCFile, 
	int numIterations, float regularizer,
	bool blnFixFirstImage, 
	int numGeodesicPts, float psiDeg, 
	float priorLambda,
	bool blnFixRotations, bool blnSaveProjections,
	const float startSDForAnnealing, 
	const float minSDForAnnealing, 
	const float epsilon)
 {
	mintNumIterations = 0 ; 
	mdblRegularizer = regularizer ; 
	mdblPriorLambda = priorLambda ; 
	mblnFixImage1 = blnFixFirstImage ; 
	mblnFixRotations = blnFixRotations ; 
	mblnSaveProjections = blnSaveProjections ; 

	std::cerr.precision(20) ; 
    	lbfgsfloatval_t fx;
	std::vector<RotationMatrix> vectRotMatrices ; 
	if (rotationFile.length() != 0)
	{
		std::cerr<<"Loading Rotation matrices..."<<std::flush ; 
		LoadRotationMatrices(rotationFile, vectRotMatrices) ; 
		std::cerr<<"Done."<<std::endl ; 
	}
	else
	{
		std::cerr<<"Creating Geodesic Rotation matrices..."<<std::flush ; 
		CreateGeodesicGridOfRotations(numGeodesicPts, psiDeg, vectRotMatrices) ; 
		std::cerr<<"Done."<<std::endl ; 
	}

	std::cerr<<"Loading projections..."<<std::flush ; 
	mobjData = new CryoData::ProjectionData(mrcFile) ;
	mobjData->Rescale() ; 
	std::cerr<<"Done."<<std::endl ; 
	std::cerr<<"Creating mobjProbabilityModel data structures..."<<std::flush ; 
	mobjProbabilityModel = new TrilinearGridModelProbabilityComputer(vectRotMatrices, mobjData->GetNumX(),
		mobjData->GetNumX()) ; 
	std::cerr<<"Done."<<std::endl ; 
	//mobjProbabilityModel->GenerateSphericalModel() ; 

	if (startModelMRCFile.length() != 0)
	{
		CryoData::ProjectionData startModel(startModelMRCFile, true) ; 
		mobjProbabilityModel->SetModelBetas(startModel.GetNumX(), (const float ***)startModel.GetDataPtr(), 1, 2, 0) ; 
	}
	else
	{
		mobjProbabilityModel->GenerateRandomModel() ; 
		//mobjProbabilityModel->GenerateSphericalModel() ; 
	}


	int N = mobjData->GetNumX()*mobjData->GetNumX()*mobjData->GetNumX() ;
	m_x = lbfgs_malloc(N);

	mobjProbabilityModel->CopyBetas((double *&)m_x) ; 
    /*
        Start the L-BFGS optimization; this will invoke the callback functions
        evaluate() and progress() when necessary.
     */
    lbfgs_parameter_t param;
	/* Initialize the parameters for the L-BFGS optimization. */
	lbfgs_parameter_init(&param);
	//param.epsilon = 1e-10 ; 
	
	param.orthantwise_c = mdblRegularizer; // suggested use .0001 when no psf, and 1e-6 otherwise
	param.linesearch = LBFGS_LINESEARCH_BACKTRACKING;

	mfltNoiseSigma = startSDForAnnealing ; 

	param.m = 10 ;

	double bestNegLL = DBL_MAX ; 
	float lastSigma = 0.0 ; 
	fx = 0 ; 
	int numAnnealings = 1 ; 
	float startAnnealing = startSDForAnnealing, endAnnealing = minSDForAnnealing ; 
	if (!blnFixRotations)
	{
		while(true)
		{
			lastSigma = 0.0 ; 
			std::cerr<<"*************************************************************\n" ; 
			std::cerr<<"Annealing standard deviations and fitting model. Cycle #: "<<numAnnealings<<std::endl ; 
			std::cerr<<"*************************************************************"<<std::endl; 
			for (int iterNum = 0 ; iterNum < numIterations ; iterNum++)
			{
				float percent = 0 ; 
				if (numIterations != 1)
					percent = iterNum*1.0f/(numIterations-1) ; 
				else
					percent = 1.0f ; 
				mfltNoiseSigma = endAnnealing 
					+ (startAnnealing-endAnnealing) * (1.0f-pow(percent,0.1f)) ; 
				if (mfltNoiseSigma < lastSigma)
					mfltNoiseSigma = lastSigma ; 

				param.max_iterations = 10 ; 
				param.max_linesearch = 10 ; 
				int ret = lbfgs(N, m_x, &fx, _evaluate, _progress, this, &param);
				//printf("L-BFGS optimization terminated with status code = %d\n", ret);
				//printf("  fx = %f, x[0] = %f, x[1] = %f\n", fx, m_x[0], m_x[1]);
				mobjProbabilityModel->SetModelBetas((const double *&)m_x) ; 
				std::stringstream ssName,ssScoresName ;
				std::string modelFileName = analysisName + "_" ;
				std::string modelScoresFileName = analysisName + "_" ;
				ssName<<std::string("annealed_")<<numAnnealings<<"_step_"<<iterNum<<std::string(".mrc") ; 
				ssScoresName<<std::string("annealed_")<<numAnnealings<<"_step_"<<iterNum<<std::string("_scores.csv") ; 
				std::string temp  ; 
				ssName>>temp ; 
				modelFileName += temp ; 
				temp.clear() ; 
				ssScoresName>>temp ; 
				modelScoresFileName += temp ; 
				mobjProbabilityModel->PrintModel(modelFileName) ; 
				mobjProbabilityModel->PrintModelScores(modelScoresFileName) ;
				lastSigma = (float) mobjProbabilityModel->GetLastErrorSigma() ; 
			}
			std::stringstream ssName,ssScoresName,ssBestProjectionName ;
			std::string modelFileName = analysisName + "_" ;
			std::string modelScoresFileName = analysisName + "_" ;
			std::string modelBestProjectionsFileName = analysisName + "_" ;
			ssName<<std::string("annealed_")<<numAnnealings<<std::string(".mrc") ; 
			ssScoresName<<std::string("annealed_")<<numAnnealings<<std::string("_scores.csv") ; 
			ssBestProjectionName<<std::string("annealed_")<<numAnnealings<<std::string("_bestProjections.mrc") ; 
			std::string temp  ; 
			ssName>>temp ; 
			modelFileName += temp ; 
			temp.clear() ; 
			ssScoresName>>temp ; 
			modelScoresFileName += temp ; 
			
			ssBestProjectionName>>temp ; 
			modelBestProjectionsFileName += temp ; 

			mobjProbabilityModel->PrintModel(modelFileName) ; 
			mobjProbabilityModel->PrintModelScores(modelScoresFileName) ;
			mobjProbabilityModel->PrintBestProjections(modelBestProjectionsFileName) ;
			mblnFixImage1 = false ; 
			if (bestNegLL - fx < epsilon)
			{
				break ; 
			}
			if (fx < bestNegLL)
			{
				bestNegLL = fx ; 
			}
			numAnnealings++ ; 
			//startAnnealing = startAnnealing *.8 ; 
			endAnnealing = lastSigma ; 
		}
		mobjProbabilityModel->PrintModel(analysisName + std::string(".mrc")) ; 
		mobjProbabilityModel->PrintModelScores(analysisName + std::string("_scores.csv")) ; 
		mobjProbabilityModel->PrintBestRotations(std::string("bestRotations.csv")) ; 
	}
	else
	{
		param.max_iterations = numIterations ; 
		param.max_linesearch = 10 ; 
		int ret = lbfgs(N, m_x, &fx, _evaluate, _progress, this, &param);
		std::cerr<<"L-BFGS optimization terminated with status code = "<<ret<<std::endl ; 
		std::cerr<<"  fx = "<<fx<<", x[0] = "<<m_x[0]<<", x[1] = "<<m_x[1]<<std::endl ; 
		mobjProbabilityModel->SetModelBetas((const double *&)m_x) ; 
		mobjProbabilityModel->PrintModel(analysisName+std::string(".mrc")) ; 
		mobjProbabilityModel->PrintModelScores(analysisName+std::string("_scores.csv")) ; 
	}

    /* Report the result. */
    
    return 0;
 }


 lbfgsfloatval_t LBFGSFitTrilinearModel::evaluate(
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g,
    const int n, const lbfgsfloatval_t step)
{
	// regularizer in trilinear model is set to 0. use the one with 
	// lbfgs
	double ll = mobjProbabilityModel->ComputeLogLikelihood(
		x, g,
		*mobjData, 0,mdblPriorLambda, 
		mblnFixImage1,mblnFixRotations, mblnSaveProjections, 
		mfltNoiseSigma) ; 		
	mintNumIterations++ ; 
	std::cerr<<" Step = "<<step<<std::endl ; 
	return -ll ; 
}

 int LBFGSFitTrilinearModel::progress(
    const lbfgsfloatval_t *x, const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx, const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm, const lbfgsfloatval_t step,
    int n, int k, int ls
    )
{
	//mobjProbabilityModel->PrintCurrentModel(k) ; 
    //printf("Iteration %d:\n", k);
    //printf(" Log likelihood fx = %f, x[0] = %f, x[1] = %f\n", fx, x[0], x[1]);
    //printf("  xnorm = %f, gnorm = %f, step = %f\n", xnorm, gnorm, step);
    //printf("\n");
    return 0;
}
