//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include "ProjectionData.h"
#include <cstdio>
#include <iostream> 
#include "RotationMatrix.h"
#include <float.h> 

namespace CryoData
{
	void SetColumnProperties(int &numX, int &numY, int &numProjections, 
		int &numElementsPerX, int &numElementsPerY, int &numElementsPerProjection, 
		const int numColumns, const int numRows, const int numSections, 
		const int mapCol, const int mapRow, const int mapSection)
	{
		if (mapCol == 1)
		{
			if (mapRow == 2)
			{
				// order: X,Y,sections
				numElementsPerX = 1; 
				numElementsPerY = numColumns ;
				numElementsPerProjection = numColumns*numRows ; 

				numX = numColumns ; 
				numY = numRows ; 
				numProjections = numSections ; 
			}
			else
			{
				// order: X,sections,Y.
				numElementsPerX = 1; 
				numElementsPerProjection =  numColumns ; 
				numElementsPerY = numColumns*numRows;

				numX = numColumns ; 
				numProjections = numRows ; 
				numY = numSections ; 
			}
		}
		else if (mapCol == 2)
		{
			if (mapRow == 1)
			{
				// order: Y,X, sections.
				numElementsPerY = 1 ;
				numElementsPerX = numColumns ; 
				numElementsPerProjection = numColumns*numRows ; 

				numY = numColumns ; 
				numX = numRows ; 
				numProjections = numSections ; 

			}
			else
			{
				// order: sections, X, Y.
				numElementsPerProjection =  1 ; 
				numElementsPerX = numColumns ; 
				numElementsPerY = numColumns*numRows ;

				numProjections = numColumns ; 
				numX = numRows ; 
				numY = numSections ; 
			}
		}
		else
		{
			if (mapRow == 1)
			{
				// order: section, X, Y
				numElementsPerProjection = 1 ;
				numElementsPerX = numColumns ; 
				numElementsPerY = numColumns*numRows ; 

				numProjections = numColumns ; 
				numX = numRows ; 
				numY = numSections ; 

			}
			else
			{
				// order: section, Y, X.
				numElementsPerProjection =  1 ; 
				numElementsPerY = numColumns ;
				numElementsPerX = numColumns*numRows ; 

				numProjections = numColumns ; 
				numY = numRows ; 
				numX = numSections ; 
			}
		}
	}

	ProjectionData::ProjectionData(const std::string &fName, bool isDensity):marrData(0),
		mblnDensity(isDensity)
	{
	// Format of mrc file from scripps: 
	//	http://ami.scripps.edu/software/mrctools/mrc_specification.php
	//LONG Word # Specification 
	//	1 NX       number of columns (fastest changing in map) 
	//	2 NY       number of rows    
	//	3 NZ       number of sections (slowest changing in map) 
	//	4 MODE     data type :
	//	 0        image : signed 8-bit bytes range -128 to 127
	//	 1        image : 16-bit halfwords
	//	 2        image : 32-bit reals
	//	 3        transform : complex 16-bit integers
	//	 4        transform : complex 32-bit reals
	//	 6        image : unsigned 16-bit range 0 to 65535

	//	5 NXSTART number of first column in map (Default = 0) 
	//	6 NYSTART number of first row in map 
	//	7 NZSTART number of first section in map 
	//	8 MX       number of intervals along X 
	//	9 MY       number of intervals along Y 
	//	10 MZ       number of intervals along Z 
	//	11-13 CELLA    cell dimensions in angstroms 

	//	14-16 CELLB    cell angles in degrees 
	//	17 MAPC     axis corresp to cols (1,2,3 for X,Y,Z) 
	//	18 MAPR     axis corresp to rows (1,2,3 for X,Y,Z) 
	//	19 MAPS     axis corresp to sections (1,2,3 for X,Y,Z) 
	//	20 DMIN     minimum density value 
	//	21 DMAX     maximum density value 
	//	22 DMEAN    mean density value 
	//	23 ISPG     space group number 0 or 1 (default=0) 
	//	24 NSYMBT   number of bytes used for symmetry data (0 or 80) 
	//	25-49 EXTRA    extra space used for anything   - 0 by default 
	//	50-52 ORIGIN   origin in X,Y,Z used for transforms 
	//	53 MAP      character string 'MAP ' to identify file type 
	//	54 MACHST   machine stamp 
	//	55 RMS      rms deviation of map from mean density 
	//	56 NLABL    number of labels being used 
	//	57-256 LABEL(20,10) 10 80-character text labels 


		FILE *fp = fopen(fName.c_str(), "rb") ; 
		if (fp == NULL)
		{
			const int BUFF_LEN = 256 ;
			char err[BUFF_LEN] ; 
			strcpy(err, "Unable to open file ") ; 
			strncat(err, fName.c_str(), std::min(BUFF_LEN-strlen(err)-1, fName.length())) ; 
			throw err ; 
		}
		int numColumns=0, numRows=0, numSections=0;
		size_t nRead = fread(&numColumns, sizeof(int), 1, fp) ; 
		nRead = fread(&numRows, sizeof(int), 1, fp) ; 
		nRead = fread(&numSections, sizeof(int), 1, fp) ; 

		int type = 0 ; 
		nRead = fread(&type, sizeof(int), 1, fp) ; 
		// just assuming they are 0 in the code below. These values are ignored for 
		// now because my feeble mind cannot comprehend why only a subset of the data
		// is to be read.
		int nxStart = 0, nyStart = 0, nzStart = 0 ; 
		nRead = fread(&nxStart, sizeof(int),1,fp) ; 
		nRead = fread(&nyStart, sizeof(int),1,fp) ; 
		nRead = fread(&nzStart, sizeof(int),1,fp) ; 
		// number of intervals. Again, my feeble mind does not comprehend why
		// these numbers would be different from the dimensionalities read in above 
		// in mintNumX(Y,Projections).
		int mx=0, my=0, mz=0 ;  
		nRead = fread(&mx, sizeof(int),1,fp) ; 
		nRead = fread(&my, sizeof(int),1,fp) ; 
		nRead = fread(&mz, sizeof(int),1,fp) ; 
		// ignoring the dimensions in angstroms because the program just works in
		// pixel space. 
		int dimXInAngs=0, dimYInAngs=0, dimZInAngs=0 ;  
		nRead = fread(&dimXInAngs, sizeof(int),1,fp) ; 
		nRead = fread(&dimYInAngs, sizeof(int),1,fp) ; 
		nRead = fread(&dimZInAngs, sizeof(int),1,fp) ; 
		// ignoring the angles in angstroms because the program just works in
		// pixel space. 
		int angsXInAngs=0, angsYInAngs=0, angsZInAngs=0 ;  
		nRead = fread(&angsXInAngs, sizeof(int),1,fp) ; 
		nRead = fread(&angsYInAngs, sizeof(int),1,fp) ; 
		nRead = fread(&angsZInAngs, sizeof(int),1,fp) ; 

		int mapCol = 0, mapRow = 0, mapSection = 0 ; 
		nRead = fread(&mapCol, sizeof(int),1,fp) ; 
		nRead = fread(&mapRow, sizeof(int),1,fp) ; 
		nRead = fread(&mapSection, sizeof(int),1,fp) ; 

		float dmin = 0.0 ; 
		float dmax = 0.0 ; 
		float davg = 0.0 ; 

		if (type == 2)
		{
			float ranges[3] ;
			nRead = fread(ranges, sizeof(float), 3, fp) ; 
			dmin = ranges[0] ; dmax = ranges[1] ; davg = ranges[2] ;
		}
		else if (type == 1)
		{
			short ranges[3] ;
			nRead = fread(ranges, sizeof(short), 3, fp) ; 
			dmin = ranges[0] ; dmax = ranges[1] ; davg = ranges[2] ;
		}
		else
		{
			throw "Unexpected type in input data file." ; 
		}

		int numElementsPerX =0, numElementsPerY=0, numElementsPerProjection=0 ;
		SetColumnProperties(mintNumX, mintNumY, mintNumProjections, 
			numElementsPerX, numElementsPerY, numElementsPerProjection, 
			numColumns, numRows, numSections, 
			mapCol, mapRow, mapSection) ; 

		int success = fseek(fp, 1024, SEEK_SET) ; 

		if (type == 2)
		{
			AllocateArrays() ; 
			float *data = new float[mintNumProjections*mintNumX*mintNumY] ; 
			size_t numRead = fread(data, sizeof(float), mintNumProjections*mintNumX*mintNumY, fp) ; 
			if(numRead != mintNumProjections*mintNumX*mintNumY)
			{
				delete [] data ; 
				throw "Could not read the correct amount of data from mrc file." ; 
			}
			for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
			{
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
						int dataIndex = xPix*numElementsPerX 
							+ yPix * numElementsPerY 
							+ projNum * numElementsPerProjection ; 

						marrData[projNum][xPix][yPix] = 
							data[dataIndex] ; 
					}
				}
			}
			delete [] data ; 
		}
		else if (type == 1)
		{
			short *data = new short[mintNumProjections*mintNumX*mintNumY ] ; 
			size_t numRead = fread(data, sizeof(short), mintNumProjections*mintNumX*mintNumY , fp) ; 
			if(numRead != mintNumX * mintNumY * mintNumProjections)
			{
				delete [] data ; 
				throw "Could not read the correct amount of data from mrc file." ; 
			}
			for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
			{
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
						int dataIndex = xPix*numElementsPerX 
							+ yPix * numElementsPerY 
							+ projNum * numElementsPerProjection ; 

						marrData[projNum][xPix][yPix] = 
							data[dataIndex] ; 
					}
				}
			}
			delete [] data ; 
		}
		fclose(fp) ; 
	}

	void ProjectionData::AllocateArrays()
	{
		if (marrData != NULL)
			DeAllocateArrays() ; 

		marrData = new float ** [mintNumProjections] ; 
		for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
		{
			marrData[projNum] = new float* [mintNumX] ; 
			for (int xPix = 0 ; xPix < mintNumX ; xPix++)
			{
				marrData[projNum][xPix] = new float [mintNumY] ; 
			}
		}
	}

	void ProjectionData::DeAllocateArrays()
	{
		if (marrData == NULL)
			return ; 
		for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
		{
			for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				delete [] marrData[projNum][xPix] ; 
			delete [] marrData[projNum] ; 
		}
		delete [] marrData ; 
		marrData = NULL ; 
	}

	ProjectionData::~ProjectionData(void)
	{
		if (marrData != NULL)
		{
			DeAllocateArrays() ; 
		}
	}

	void ProjectionData::PrintData(std::ostream &out, int projectionNum) const
	{
		int startProjNum = 0 ; 
		int endProjNum = mintNumProjections - 1; 
		if (projectionNum != -1)
		{
			startProjNum = projectionNum ; 
			endProjNum = projectionNum ; 
		}
		int numPixelsPerProjection = mintNumX * mintNumY ; 
		for (int projNum = startProjNum ; projNum <= endProjNum ; projNum++)
		{
			for (int xPix = 0 ; xPix < mintNumX ; xPix++)
			{
				for (int yPix = 0 ; yPix < mintNumY ; yPix++)
				{
					out<<marrData[projNum][xPix][yPix] ; 
					if (yPix != mintNumY-1)
						out<<"," ; 
				}
				out<<"\n" ; 
			}
		}
	}

	void ProjectionData::SetData(int numProjections, int imageSide, 
		const float *** projections, bool blnIsDensity)
	{
		mblnDensity = blnIsDensity ; 
		if (marrData != NULL && (mintNumProjections != numProjections || mintNumX != imageSide 
			|| mintNumY != imageSide))
		{
			DeAllocateArrays() ; 
		}
		mintNumProjections = numProjections ; 
		mintNumX = mintNumY = imageSide ; 
		marrData = new float ** [mintNumProjections] ; 
		if (blnIsDensity)
		{
			for (int zIndex = 0 ; zIndex < mintNumProjections ; zIndex++)
			{
				marrData[zIndex] = new float* [mintNumX] ; 
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					marrData[zIndex][xPix] = new float [mintNumY] ; 
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
						marrData[zIndex][xPix][yPix] = (float) projections[xPix][yPix][zIndex] ; 
					}
				}
			}
		}else
		{
			for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
			{
				marrData[projNum] = new float* [mintNumX] ; 
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					marrData[projNum][xPix] = new float [mintNumY] ; 
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
							marrData[projNum][xPix][yPix] = (float) projections[projNum][xPix][yPix] ; 
					}
				}
			}
		}
	}
	// Copies the data for the projections from the input.
	// when blnIsDensity is true, the data is assumed to be
	// otherwise its assumed that there are many projections
	// projections[xIndex][yIndex][zIndex], where numProjections = numZ.
	// and the data is projection by projections: i.e
	// projections[projection#][xIndex][yIndex]

	void ProjectionData::SetData(int numProjections, int imageSide, 
		const double *** projections, bool blnIsDensity)
	{
		mblnDensity = blnIsDensity ; 
		if (marrData != NULL && (mintNumProjections != numProjections || mintNumX != imageSide 
			|| mintNumY != imageSide))
		{
			DeAllocateArrays() ; 
		}
		mintNumProjections = numProjections ; 
		mintNumX = mintNumY = imageSide ; 
		marrData = new float ** [mintNumProjections] ; 

		if (blnIsDensity)
		{
			for (int zIndex = 0 ; zIndex < mintNumProjections ; zIndex++)
			{
				marrData[zIndex] = new float* [mintNumX] ; 
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					marrData[zIndex][xPix] = new float [mintNumY] ; 
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
						marrData[zIndex][xPix][yPix] = (float) projections[xPix][yPix][zIndex] ; 
					}
				}
			}
		}else
		{
			for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
			{
				marrData[projNum] = new float* [mintNumX] ; 
				for (int xPix = 0 ; xPix < mintNumX ; xPix++)
				{
					marrData[projNum][xPix] = new float [mintNumY] ; 
					for (int yPix = 0 ; yPix < mintNumY ; yPix++)
					{
						marrData[projNum][xPix][yPix] = (float) projections[projNum][xPix][yPix] ; 
					}
				}
			}
		}
	}

	void ProjectionData::WriteMRC(const std::string &outFile)
	{
		FILE *fp = fopen(outFile.c_str(), "wb") ; 
		if (fp == NULL)
		{
			const int BUFF_LEN = 256 ;
			char err[BUFF_LEN] ; 
			strcpy(err, "Unable to open file for writing: ") ; 
			strncat(err, outFile.c_str(), std::min(BUFF_LEN-strlen(err)-1, outFile.length())) ; 
			throw err ; 
		}
		int header [256] ; 
		for (int i = 0 ; i < 256 ; i++)
			header[i] = 0 ; 
		header[0] = (int)mintNumX ; 
		header[1] = (int)mintNumY ; 
		header[2] = (int)mintNumProjections ; 
		header[3] = 2 ; // signifying that data is floats. 

		// DIMENSION OF DATA (NX, NY, NZ) and TYPE..
		int status = fwrite(header, sizeof(int), 4, fp) ; 
		// NXSTART, NYSTART, NZSTART..
		status=fwrite(header,sizeof(int), 3, fp);
		// MX,MY,MZ: Number of intervals along X,Y,Z.
		int Z[] = {0,0,0} ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// CELLA cell dimensions in Angstroms.
		Z[0] = Z[1] = Z[2] = 1 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// CELLB cell angles in Angstroms.
		Z[0] = Z[1] = Z[2] = 0 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// MAPC axis corresponds fo columns
		Z[0] = 1 ; Z[1] = 2 ; Z[2] = 3 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// MIN MAX and AVG. 
		float minV=FLT_MAX, maxV=-FLT_MAX, avgV = 0 ; 
		for (int imageNum = 0 ; imageNum < mintNumProjections ; imageNum++)
			for (int xIndex = 0 ; xIndex < mintNumX ; xIndex++)
				for (int yIndex = 0 ; yIndex < mintNumX ; yIndex++)
				{
					float val = marrData[imageNum][xIndex][yIndex] ; 
					if (val < minV)
						minV = val ; 
					if (val > maxV)
						maxV = val ; 
					avgV += val ; 
				}
		avgV /= (mintNumProjections*mintNumX*mintNumY) ; 
		status=fwrite(&minV,sizeof(float), 1, fp);
		status=fwrite(&maxV,sizeof(float), 1, fp);
		status=fwrite(&avgV,sizeof(float), 1, fp);
		// Unknown stuff. Write zeros 
		status=fwrite(&header[4],sizeof(int), 30, fp);
		// MAP type. 
		int decForMap = 542130509 ; 
		status=fwrite(&decForMap,sizeof(int), 1, fp);
		// remaining unknown stuff. 203 words. Again write 0s.
		status=fwrite(&header[4],sizeof(int), 203, fp);

		// and now, time for data. 
		// but our data is spread all over the place. Consolidate to one
		// continuous array. 
		float *data = new float [mintNumX*mintNumY*mintNumProjections] ; 
		int index = 0 ; 
		// xIndex changes the fasted by the column map above, then yIndex and 
		// lastly zIndex.
		for (int imageNum = 0 ; imageNum < mintNumProjections ; imageNum++)
			for (int yIndex = 0 ; yIndex < mintNumY ; yIndex++)
				for (int xIndex = 0 ; xIndex < mintNumX ; xIndex++)
					data[index++] = marrData[imageNum][xIndex][yIndex] ; 

		status = fwrite(data, sizeof(float), mintNumX*mintNumY*mintNumProjections, fp) ; 
		delete [] data ;
		fclose(fp) ; 
	}
	FILE* ProjectionData::OpenMRCForWritingProjections(
		const std::string &outFile, 
		int numX, int numY, int numZ)
	{
		FILE *fp = fopen(outFile.c_str(), "wb") ; 
		if (fp == NULL)
		{
			const int BUFF_LEN = 256 ;
			char err[BUFF_LEN] ; 
			strcpy(err, "Unable to open file for writing: ") ; 
			strncat(err, outFile.c_str(), std::min(BUFF_LEN-strlen(err)-1, outFile.length())) ; 
			throw err ; 
		}
		int header [256] ; 
		for (int i = 0 ; i < 256 ; i++)
			header[i] = 0 ; 
		header[0] = (int)numX ; 
		header[1] = (int)numY ; 
		header[2] = (int)numZ ; 
		header[3] = 2 ; // signifying that data is floats. 

		// DIMENSION OF DATA (NX, NY, NZ) and TYPE..
		int status = fwrite(header, sizeof(int), 4, fp) ; 
		// NXSTART, NYSTART, NZSTART..
		status=fwrite(header,sizeof(int), 3, fp);
		// MX,MY,MZ: Number of intervals along X,Y,Z.
		int Z[] = {0,0,0} ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// CELLA cell dimensions in Angstroms.
		Z[0] = Z[1] = Z[2] = 1 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// CELLB cell angles in Angstroms.
		Z[0] = Z[1] = Z[2] = 0 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// MAPC axis corresponds fo columns
		Z[0] = 1 ; Z[1] = 2 ; Z[2] = 3 ; 
		status=fwrite(Z,sizeof(int), 3, fp);
		// MIN MAX and AVG. 
		float minV=0, maxV=0, avgV = 0 ; 
		status=fwrite(&minV,sizeof(float), 1, fp);
		status=fwrite(&maxV,sizeof(float), 1, fp);
		status=fwrite(&avgV,sizeof(float), 1, fp);
		// Unknown stuff. Write zeros 
		status=fwrite(&header[4],sizeof(int), 30, fp);
		// MAP type. 
		int decForMap = 542130509 ; 
		status=fwrite(&decForMap,sizeof(int), 1, fp);
		// remaining unknown stuff. 203 words. Again write 0s.
		status=fwrite(&header[4],sizeof(int), 203, fp);
		return fp ; 
	}

	void ProjectionData::AddProjectionToOpenFile(
		FILE *fp, int numX, int numY, const float **(&projection))
	{
		float *data = new float [numX*numY] ; 
		int index = 0 ; 
		for (int xIndex = 0 ; xIndex < numX ; xIndex++)
			for (int yIndex = 0 ; yIndex < numX ; yIndex++)
				data[index++] = projection[xIndex][yIndex] ; 

		int status = fwrite(data, sizeof(float), numX*numY, fp) ; 
		delete [] data ;
	}
	void ProjectionData::AddProjectionToOpenFile(
		FILE *fp, int numX, int numY, const double **(&projection))
	{
		float *data = new float [numX*numY] ; 
		int index = 0 ; 
		for (int xIndex = 0 ; xIndex < numX ; xIndex++)
			for (int yIndex = 0 ; yIndex < numX ; yIndex++)
				data[index++] = (float) projection[xIndex][yIndex] ; 

		int status = fwrite(data, sizeof(float), numX*numY, fp) ; 
		delete [] data ;
	}
	void ProjectionData::Rescale()
	{
		mfltMaxValue = -FLT_MAX ; 
		mfltMinValue = FLT_MAX ; 
		float MAX_VAL = 1 ; 
		for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
		{
			for (int x = 0 ; x < mintNumX ; x++)
			{
				for (int y = 0 ; y < mintNumY ; y++)
				{
					if (mfltMaxValue < marrData[projNum][x][y])
					{
						mfltMaxValue = marrData[projNum][x][y] ; 
					}
					if (mfltMinValue > marrData[projNum][x][y])
					{
						mfltMinValue = marrData[projNum][x][y] ; 
					}
				}
			}
		}
		for (int projNum = 0 ; projNum < mintNumProjections ; projNum++)
		{
			for (int x = 0 ; x < mintNumX ; x++)
			{
				for (int y = 0 ; y < mintNumY ; y++)
				{
					marrData[projNum][x][y] = MAX_VAL * marrData[projNum][x][y]/mfltMaxValue ; 
				}
			}
		}
		mfltMinValue = 0.0f ; 
		mfltMaxValue = MAX_VAL ; 
	}
}
