//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include "RotationMatrix.h"
#include <time.h> 
#include <algorithm>
#include <cmath>
#include <fstream> 

const double PI = 4*atan(1.0) ; 

RotationMatrix::RotationMatrix(void)
{
}

RotationMatrix::~RotationMatrix(void)
{
}

RotationMatrix::RotationMatrix(float psi, float theta, float phi)
{
	SetFromEulerYZY(psi, theta, phi) ; 
}

RotationMatrix::RotationMatrix(float a, float b, float c, float d)
{
	SetFromQuaternion(a,b,c,d) ; 
}

RotationMatrix::RotationMatrix(const float (&rots)[3][3])
{
	SetMatrixEntries(rots) ; 
}

void RotationMatrix::SetFromQuaternion(float a, float b, float c, float d)
{
}

void RotationMatrix::SetFromEulerYZY(float psi, float theta, float phi)
{
}

void RotationMatrix::SetMatrixEntries(const float (&rots)[3][3])
{
	memcpy(marrData[0], rots[0], 3*sizeof(float)) ; 
	memcpy(marrData[1], rots[1], 3*sizeof(float)) ; 
	memcpy(marrData[2], rots[2], 3*sizeof(float)) ; 
}


void RotationMatrix::CreateGeodesicRotations(const int numDivisions, const float numAnglesPerPoint, 
		std::vector<RotationMatrix> &vectRotationMatrix)
{
	std::vector<Point3D> vectPoints ; 
	CreateGeodesicPoints(vectPoints, numDivisions) ; 
	int numPts = vectPoints.size() ; 
	std::vector<float> vectAngles ;
	for (float angle = 0 ; angle < 360 ; angle += numAnglesPerPoint)
		vectAngles.push_back(angle/360) ; 

	for (int ptNum =  0 ; ptNum < numPts ; ptNum++)
	{
		Point3D point = vectPoints[ptNum] ; 
		float x3 = float (0.5*(1+point.mdblZ)) ; 
        float twoPiX2 = (float) (atan2(point.mdblY, point.mdblX) + PI) ; 
        float x2 =  float (twoPiX2/(2*PI)) ; 
		for (std::vector<float>::iterator iter = vectAngles.begin() ; iter != vectAngles.end() ; 
			iter++)
		{
			RotationMatrix mat = CreateRotationMatrixFromRandPts(*iter, x2, x3) ; 
			vectRotationMatrix.push_back(mat) ; 
			Point p = mat.Rotate(GridIntersection::Z_DIRECTION) ; 
		}
	}
}

void RotationMatrix::CreateGeodesicPoints(std::vector<Point3D> &vectPoints, int numGeodesicDivisions)
{
	double t1 = 2*PI/5;
	double t2 = PI/10;
	double t3 = -3*PI/10;
	double t4 = PI/5;

	double R = 1.0 ;
	double side = 2*R*sin(t4);
	double S = side;

	double H = cos(t4)*R;
	double Cx = R * cos(t2);
	double Cy = R * sin(t2);

	double H1 = sqrt(side*side - R*R);
	double H2 = sqrt((H+R)*(H+R) - H*H);
	double Z2 = (H2 - H1)/2;
	double Z1 = Z2 + H1;

	vectPoints.resize(12); 
	vectPoints[0] = Point3D(   0,  0, Z1);
	vectPoints[1] = Point3D(   0,  R, Z2);
	vectPoints[2] = Point3D(  Cx, Cy, Z2);
	vectPoints[3] = Point3D( S/2, -H, Z2);
	vectPoints[4] = Point3D(-S/2, -H, Z2);
	vectPoints[5] = Point3D( -Cx, Cy, Z2);
	vectPoints[6] = Point3D(   0, -R,-Z2);
	vectPoints[7] = Point3D( -Cx,-Cy,-Z2);
	vectPoints[8] = Point3D(-S/2,  H,-Z2);
	vectPoints[9] = Point3D( S/2,  H,-Z2);
	vectPoints[10] = Point3D( Cx,-Cy,-Z2);
	vectPoints[11] = Point3D(  0,  0,-Z1);

	// now create the faces
	std::vector<Face3> face ; 
	face.push_back(Face3(0,1,2)) ; 
	face.push_back(Face3(0,2,3)) ; 
	face.push_back(Face3(0,3,4)) ; 
	face.push_back(Face3(0,4,5)) ; 
	face.push_back(Face3(0,5,1)) ; 
	face.push_back(Face3(1,9,2)) ; 
	face.push_back(Face3(2,9,10)) ; 
	face.push_back(Face3(2,10,3)) ; 
	face.push_back(Face3(3,10,6)) ; 
	face.push_back(Face3(3,6,4)) ; 
	face.push_back(Face3(4,6,7)) ; 
	face.push_back(Face3(4,7,5)) ; 
	face.push_back(Face3(5,7,8)) ; 
	face.push_back(Face3(5,8,1)) ; 
	face.push_back(Face3(1,8,9)) ; 
	face.push_back(Face3(9,11,10)) ; 
	face.push_back(Face3(10,11,6)) ; 
	face.push_back(Face3(6,11,7)) ; 
	face.push_back(Face3(7,11,8)) ; 
	face.push_back(Face3(8,11,9)) ; 

	// now we have our icosahedron
	// let's extend it to a geodesic sphere
	// subdivide each face
	std::vector<Face3> facesGS = face;
	for (int subdiv=1; subdiv<=numGeodesicDivisions; ++subdiv)
	{
		std::vector<Face3> facesT = facesGS;    // temp face list (previous subdivision)
		facesGS.clear();
		//vectPoints.clear();
		for (size_t i=0; i<facesT.size(); ++i)
		{
			Face3 &f = facesT[i];

			// create the new points
			Point3D newV1 = 0.5*(vectPoints[f.mintID[0]] + vectPoints[f.mintID[1]]);
			Point3D newV2 = 0.5*(vectPoints[f.mintID[1]] + vectPoints[f.mintID[2]]);
			Point3D newV3 = 0.5*(vectPoints[f.mintID[2]] + vectPoints[f.mintID[0]]);

			// create the 3 new faces
			int id0 = f.mintID[0];
			int id1 = f.mintID[1];
			int id2 = f.mintID[2];
			int idN1 = (int)vectPoints.size();
			std::vector<Point3D>::iterator iter = std::find(vectPoints.begin(), vectPoints.end(), newV1) ;
			if (iter == vectPoints.end())
			{
				vectPoints.push_back(newV1) ; 
			}
			else
			{
				idN1 = (int) (iter - vectPoints.begin()) ; 
			}
			
			int idN2 = (int)vectPoints.size();
			iter = std::find(vectPoints.begin(), vectPoints.end(), newV2) ;
			if (iter == vectPoints.end())
			{
				vectPoints.push_back(newV2) ; 
				idN2 = vectPoints.size()-1 ; 
			}
			else
			{
				idN2 = (int) (iter - vectPoints.begin()) ; 
			}

			int idN3 = (int)vectPoints.size();
			iter = std::find(vectPoints.begin(), vectPoints.end(), newV3) ;
			if (iter == vectPoints.end())
			{
				vectPoints.push_back(newV3) ; 
				idN3 = vectPoints.size()-1 ; 
			}
			else
			{
				idN3 = (int) (iter - vectPoints.begin()) ; 
			}


			facesGS.push_back(Face3(id0, idN1, idN3)) ; 
			facesGS.push_back(Face3(idN1, id1, idN2)) ; 
			facesGS.push_back(Face3(idN2, id2, idN3)) ; 
			facesGS.push_back(Face3(idN1, idN2, idN3)) ; 
		}
	}

	for (size_t i=0; i<vectPoints.size(); ++i)
	{
		Point3D &v = vectPoints[i];
		v.Normalize() ; 
	}
}


RotationMatrix RotationMatrix::CreateRotationMatrixFromRandPts(float x1, float x2, float x3)
{
	float test = cos((float)PI/4.0f) ; 
	float test2 = cos(45.0f) ; 
	float cosX2 = (float) cos(2*PI*x2) ; 
	float sinX2 = (float) sin(2*PI*x2) ; 
	float cosX1 = (float) cos(2*PI*x1) ; 
	float sinX1 = (float) sin(2*PI*x1) ; 

	if (fabs(cosX2) == 1)
		sinX2 = 0 ; 
	if (fabs(sinX2) == 1)
		cosX2 = 0 ; 
	if (fabs(cosX1) == 1)
		sinX1 = 0 ; 
	if (fabs(sinX1) == 1)
		cosX1 = 0 ; 

	float R1 [3][3] ; 

	float v[3] ; 
	v[0] = sqrt(x3)*cosX2 ; 
	v[1] = sqrt(x3)*sinX2 ; 
	v[2] = sqrt(1-x3) ; 

	for (int i = 0 ; i < 3 ; i++)
		for (int j = 0 ; j < 3 ; j++)
			R1[i][j] = -2*v[i] * v[j] ; 
	R1[0][0] += 1.0f ; 
	R1[1][1] += 1.0f ; 
	R1[2][2] += 1.0f ; 

	float R2 [3][3] ; 
	R2[0][0] = cosX1 ; R2[0][1] = sinX1 ; R2[0][2] = 0 ; 
	R2[1][0] = -sinX1 ; R2[1][1] = cosX1 ; R2[1][2] = 0 ; 
	R2[2][0] = 0 ; R2[2][1] = 0 ; R2[2][2] = 1 ; 

	float R3[3][3] ; 
	for (int i = 0 ; i < 3 ; i++)
		for (int j = 0 ; j < 3 ; j++)
		{
			R3[i][j] = 0 ; 
			for (int k = 0 ; k < 3 ; k++)
			{
				R3[i][j] -=  R1[i][k] * R2[k][j] ; 
			}
		}
	return RotationMatrix(R3) ; 
}

RotationMatrix RotationMatrix::GenerateRandomRotationMatrix()
{
	//float rotMatrix [][3] = {
	//{0.26623686f,-0.05650232f,-0.96225019f},
	//{-0.82523752f,0.50249834f,-0.25783416f},
	//{0.49809735f,0.86272992f,0.08715574f}} ; 

	// generate one randomly. 
	srand(clock()) ; 
	float x1 = (float) (rand()*1.0f)/RAND_MAX ; 
	float x2 = (float) (rand()*1.0f)/RAND_MAX ; 
	float x3 = (float) (rand()*1.0f)/RAND_MAX ; 
	return CreateRotationMatrixFromRandPts(x1, x2, x3) ; 
}

std::ostream& operator<<(std::ostream& out, const RotationMatrix &mat)
{
	out<<mat.marrData[0][0]<<","<<mat.marrData[0][1]<<","<<mat.marrData[0][2]<<"," ; 
	out<<mat.marrData[1][0]<<","<<mat.marrData[1][1]<<","<<mat.marrData[1][2]<<"," ; 
	out<<mat.marrData[2][0]<<","<<mat.marrData[2][1]<<","<<mat.marrData[2][2] ; 
	return out ; 
}
