//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly

#pragma once
#include <cstring>
#include "GridIntersection.h" 


class RotationMatrix
{
	float marrData [3][3] ; 
	class Point3D
	{
	public:
		double mdblX ; 
		double mdblY ; 
		double mdblZ ; 
		Point3D(): mdblX(0), mdblY(0), mdblZ(0) {} ; 
		Point3D(double x, double y, double z): mdblX(x), mdblY(y), mdblZ(z) {} ; 
		~Point3D() {} ; 
		Point3D(const Point3D &a) { mdblX = a.mdblX ; mdblY = a.mdblY ; mdblZ = a.mdblZ ; } ; 
		Point3D& operator =(const Point3D &a) 
		{ 
			if (this == &a) return *this ;  
			mdblX = a.mdblX ; mdblY = a.mdblY ; mdblZ = a.mdblZ ;  
			return *this ;
		} ; 
		friend Point3D operator *(float c, const Point3D &a) 
		{ 
			return Point3D(c*a.mdblX, c*a.mdblY, c*a.mdblZ) ;  
		} ; 
		friend Point3D operator +(const Point3D &a, const Point3D &b) 
		{
			return Point3D(a.mdblX + b.mdblX, a.mdblY + b.mdblY, a.mdblZ + b.mdblZ) ;
		}
		friend bool operator ==(const Point3D &a, const Point3D &b) 
		{
			return a.mdblX == b.mdblX && a.mdblY == b.mdblY && a.mdblZ == b.mdblZ ;
		}
		double Norm() { return sqrt(mdblX * mdblX + mdblY * mdblY + mdblZ * mdblZ) ; } 
		void Normalize() 
		{ 
			double norm = Norm() ; 
			mdblX /= norm ; 
			mdblY /= norm ; 
			mdblZ /= norm ; 
		} 
	} ; 
	class Face3
	{
	public:
		int mintID[3] ; 
		Face3() { mintID[0] = -1 ; mintID[1] = -1 ; mintID[2] = -1 ; } ; 
		Face3(int id1, int id2, int id3) { mintID[0] = id1 ; mintID[1] = id2 ; mintID[2] = id3 ; } ; 
	} ; 

	static void CreateGeodesicPoints(std::vector<Point3D> &vectPoints, int numGeodesicDivisions) ; 
public:
	friend std::ostream& operator<<(std::ostream& out, const RotationMatrix &mat) ;

	RotationMatrix(void);
	~RotationMatrix(void);
	RotationMatrix(float psi, float theta, float phi) ; 
	RotationMatrix(float a, float b, float c, float d) ; 
	RotationMatrix(const float (&rots)[3][3]) ; 
	RotationMatrix(const RotationMatrix &rot)
	{
		rot.GetData(this->marrData) ; 
	}

	RotationMatrix& operator=(const RotationMatrix &rot)
	{
		if (this == &rot)
			return *this ; 

		rot.GetData(this->marrData) ; 
		return *this ; 
	}

	void SetFromQuaternion(float a, float b, float c, float d) ; 
	void SetFromEulerYZY(float psi, float theta, float phi) ; 
	void SetMatrixEntries(const float (&rots)[3][3]) ; 
	inline Point Rotate(const Point &pt) const 
	{ 
		// return R*pt
		Point rotatedPt(
			marrData[0][0] * pt.mfltX + marrData[0][1] * pt.mfltY + marrData[0][2] * pt.mfltZ,
			marrData[1][0] * pt.mfltX + marrData[1][1] * pt.mfltY + marrData[1][2] * pt.mfltZ,
			marrData[2][0] * pt.mfltX + marrData[2][1] * pt.mfltY + marrData[2][2] * pt.mfltZ
			) ; 
		return rotatedPt ;
	}
	inline Point InverseRotate(const Point &pt) const 
	{ 
		// returns R'*pt, which is R^-1 * pt because R is orthogonal.
		Point rotatedPt(
			marrData[0][0] * pt.mfltX + marrData[1][0] * pt.mfltY + marrData[2][0] * pt.mfltZ,
			marrData[0][1] * pt.mfltX + marrData[1][1] * pt.mfltY + marrData[2][1] * pt.mfltZ,
			marrData[0][2] * pt.mfltX + marrData[1][2] * pt.mfltY + marrData[2][2] * pt.mfltZ
			) ; 
		return rotatedPt ;
	}
	inline void GetData(float (&matrix)[3][3]) const
	{
		memcpy(matrix[0], marrData[0], 3*sizeof(float)) ; 
		memcpy(matrix[1], marrData[1], 3*sizeof(float)) ; 
		memcpy(matrix[2], marrData[2], 3*sizeof(float)) ; 
	} ;

	static void CreateGeodesicRotations(const int numDivisions, const float numAnglesPerPoint, 
		std::vector<RotationMatrix> &vectRotationMatrix) ; 
	static RotationMatrix CreateRotationMatrixFromRandPts(float x1, float x2, float x3) ; 
	static RotationMatrix GenerateRandomRotationMatrix() ; 

};

