//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include <cstdlib>
#include "TrilinearGridContribution.h"
#include "GridIntersection.h"
#include "RotationMatrix.h"

const int MIN_GRID_CONTRIB_SIZE_RESERVED = 1024 ; 

void TrilinearGridContribution::ComputePointContributions(const unsigned char imageXIndex, 
		const unsigned char imageYIndex, 
		const unsigned char gridSide, 
		const std::vector<Point> &vectPoints, 
		std::vector<TrilinearGridContribution> &vectGridContributions,
		RotationMatrix &rotMatrix)
{
	int numPoints = vectPoints.size() ;
	if (numPoints == 0)
		return ; 
	if (vectGridContributions.capacity() < MIN_GRID_CONTRIB_SIZE_RESERVED)
	{
		vectGridContributions.reserve(MIN_GRID_CONTRIB_SIZE_RESERVED) ; 
	}

	const float TOL = (float) 1e-7 ; 
	Point zDirection = rotMatrix.Rotate(GridIntersection::Z_DIRECTION) ; 
	//Point zDirection = rotMatrix.InverseRotate(GridIntersection::Z_DIRECTION) ; 

	bool isZeroX, isZeroY, isZeroZ ; 
	isZeroX = isZeroY = isZeroZ = false ; 

	bool increasingX, increasingY, increasingZ ; 
	increasingX = increasingY = increasingZ = true ; 

	if (zDirection.mfltX <= 0)
		increasingX = false ; 
	if (zDirection.mfltX == 0)
		isZeroX = true ; 

	if (zDirection.mfltY <= 0)
		increasingY = false ; 
	if (zDirection.mfltY == 0)
		isZeroY = true ; 

	if (zDirection.mfltZ <= 0)
		increasingZ = false ; 
	if (zDirection.mfltZ == 0)
		isZeroZ = true ; 

	const Point &pZero = vectPoints[0] ; 
	int xIndexStart = (int) ((pZero.mfltX + 1.0f)*(gridSide-1) * .5) ; 
	int yIndexStart = (int) ((pZero.mfltY + 1.0f)*(gridSide-1) *.5) ; 
	int zIndexStart = (int) ((pZero.mfltZ + 1.0f)*(gridSide-1) * .5); 
	if (fabs(pZero.mfltX-(float)(((xIndexStart+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
		xIndexStart++ ; 
	if (fabs(pZero.mfltY-(float)(((yIndexStart+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
		yIndexStart++ ; 
	if (fabs(pZero.mfltZ-(float)(((zIndexStart+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
		zIndexStart++ ; 

	int xIndexNext, yIndexNext, zIndexNext  ; 	

	unsigned int contribIndex = 0 ; 
	int x1, x2, y1, y2, z1, z2 ; 
	for (int i = 1 ; i < numPoints ; i++)
	{
		const Point &pStart = vectPoints[i-1] ; 
		const Point &pNext = vectPoints[i] ; 

		xIndexNext = (int)((pNext.mfltX + 1.0f)*(gridSide-1) * .5); 
		yIndexNext = (int)((pNext.mfltY + 1.0f)*(gridSide-1) * .5); 
		zIndexNext = (int)((pNext.mfltZ + 1.0f)*(gridSide-1) * .5); 

		if (fabs(pNext.mfltX-(float)(((xIndexNext+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
			xIndexNext++ ; 
		if (fabs(pNext.mfltY-(float)(((yIndexNext+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
			yIndexNext++ ; 
		if (fabs(pNext.mfltZ-(float)(((zIndexNext+1) * 2.0f)/(gridSide-1)-1.0f)) < TOL)
			zIndexNext++ ; 

		if(abs(xIndexNext-xIndexStart) == 2)
		{
			// means that one of these indices calculations went too far.
			if (fabs((xIndexStart * 2.0f)/(gridSide-1)-1.0f-pStart.mfltX) 
				> fabs((xIndexNext * 2.0f)/(gridSide-1)-1.0f-pNext.mfltX))
			{
				//xIndexStart went too far. 
				if (increasingX)
					xIndexStart++ ; 
				else
					xIndexStart-- ; 
			}
			else
			{
				//xIndexEnd went too far. 
				if (increasingX)
					xIndexNext-- ; 
				else
					xIndexNext++ ; 
			}

		}
		if(abs(yIndexNext-yIndexStart) == 2)
		{
			// means that one of these indices calculations went too far.
			if (fabs((yIndexStart * 2.0f)/(gridSide-1)-1.0f-pStart.mfltY) 
				> fabs((yIndexNext * 2.0f)/(gridSide-1)-1.0f-pNext.mfltY))
			{
				//yIndexStart went too far. 
				if (increasingY)
					yIndexStart++ ; 
				else
					yIndexStart-- ; 
			}
			else
			{
				//xIndexEnd went too far. 
				if (increasingY)
					yIndexNext-- ; 
				else
					yIndexNext++ ; 
			}

		}
		if(abs(zIndexNext-zIndexStart) == 2)
		{
			// means that one of these indices calculations went too far.
			if (fabs((zIndexStart * 2.0f)/(gridSide-1)-1.0f-pStart.mfltZ) 
				> fabs((zIndexNext * 2.0f)/(gridSide-1)-1.0f-pNext.mfltZ))
			{
				//zIndexStart went too far. 
				if (increasingZ)
					zIndexStart++ ; 
				else
					zIndexStart-- ; 
			}
			else
			{
				//xIndexEnd went too far. 
				if (increasingZ)
					zIndexNext-- ; 
				else
					zIndexNext++ ; 
			}

		}

		// if both xIndexStart and xIndexNext are the same, we are staying in the same xVoxel.
		// this voxel is bounded by xIndexStart and xIndexStart + 1 on the xside. 
		// if xIndexNext = xIndexStart + 1 then voxel is contributions from xIndexStart and xIndexStart+1
		// if xIndexNext = xIndexStart - 1 then voxel is contributions from xIndexStart and xIndexStart-1
		if (xIndexStart == xIndexNext) 
		{
			// either they are the same or object passes through plane. 
			if(isZeroX)
			{
				if (pNext.mfltX == (float)((xIndexStart * 2.0f)/(gridSide-1)-1.0f))
					x1 = x2 = xIndexStart ; 
				else
				{
					x1 = xIndexStart ; 
					x2 = x1 + 1 ; 
				}
			}
			else
			{
				x1 = xIndexStart ; 
				x2 = xIndexStart + 1 ; 
			}
		}
		else
		{
			// start point is on x boundary, and we are decreasing. 
			x1 = xIndexStart ; 
			x2 = xIndexNext ; 
		}

		if (yIndexStart == yIndexNext) 
		{
			// either they are the same or object passes through plane. 
			if(isZeroY)
			{
				if (pNext.mfltY == (float)((yIndexStart * 2.0f)/(gridSide-1)-1.0f))
					y1 = y2 = yIndexStart ; 
				else
				{
					y1 = yIndexStart ; 
					y2 = y1 + 1 ; 
				}
			}
			else
			{
				y1 = yIndexStart ; 
				y2 = yIndexStart + 1 ; 
			}
		}
		else
		{
			// start point is on x boundary, and we are decreasing. 
			y1 = yIndexStart ; 
			y2 = yIndexNext ; 
		}

		if (zIndexStart == zIndexNext) 
		{
			// either they are the same or object passes through plane. 
			if(isZeroZ)
			{
				if (pNext.mfltZ == (float)((zIndexStart * 2.0f)/(gridSide-1)-1.0f))
					z1 = z2 = zIndexStart ; 
				else
				{
					z1 = zIndexStart ; 
					z2 = z1 + 1 ; 
				}
			}
			else
			{
				z1 = zIndexStart ; 
				z2 = zIndexStart + 1 ; 
			}
		}
		else
		{
			// start point is on x boundary, and we are decreasing. 
			z1 = zIndexStart ; 
			z2 = zIndexNext ; 
		}


		Point diff = pNext - pStart ; 
		float rayLength = diff.Norm() ; 
		//time_t start = clock() ; 
		if (z1 != z2 && x1 != x2 && y1 != y2)
		{
			// add contributions of eight edges. 
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, imageXIndex,
				imageYIndex, contrib)) ; 

			if (z2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
									(y1*2.0f)/(gridSide-1)-1.0f,
									(z2*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (y2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
									(y2*2.0f)/(gridSide-1)-1.0f,
									(z1*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (y2 < gridSide && z2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
									(y2*2.0f)/(gridSide-1)-1.0f,
									(z2*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
									(y1*2.0f)/(gridSide-1)-1.0f,
									(z1*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide && z2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
									(y1*2.0f)/(gridSide-1)-1.0f,
									(z2*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide && y2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
									(y2*2.0f)/(gridSide-1)-1.0f,
									(z1*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y2, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide && y2 < gridSide && z2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
									(y2*2.0f)/(gridSide-1)-1.0f,
									(z2*2.0f)/(gridSide-1)-1.0f, 
									pStart, pNext, 2.0f/(gridSide-1),
									rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y2, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (z1 != z2 && x1 != x2)
		{
			// y coordinate is the same. 
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, 
				imageXIndex, imageYIndex, contrib)) ; 

			if (z2 < gridSide) 
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z2*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide && z2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z2*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (z1 != z2 && y1 != y2)
		{
			// x coordinate is the same. 
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, 
				imageXIndex, imageYIndex, contrib)) ; 

			if (z2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z2*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (y2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y2*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (y2 < gridSide && z2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y2*2.0f)/(gridSide-1)-1.0f,
								(z2*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (x1 != x2 && y1 != y2)
		{
			// z coordinate is the same. 
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, 
				imageXIndex, imageYIndex, contrib)) ; 

			if (y2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y2*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}

			if (x2 < gridSide && y2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
								(y2*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y2, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (z1 != z2)
		{
			// x and y are the same
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, 
					imageXIndex, imageYIndex, contrib)) ; 

			if (z2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z2*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z2, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (x1 != x2)
		{
			// z and y are the same
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1, 
				imageXIndex, imageYIndex, contrib)) ; 

			if (x2 < gridSide)
			{
				contrib = GetContributions((x2*2.0f)/(gridSide-1)-1.0f, 
								(y1*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x2, y1, z1, 
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else if (y1 != y2)
		{
			// z and x are the same
			double contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
							(y1*2.0f)/(gridSide-1)-1.0f,
							(z1*2.0f)/(gridSide-1)-1.0f, 
							pStart, pNext, 2.0f/(gridSide-1),
							rayLength) ; 
			vectGridContributions.push_back(TrilinearGridContribution(x1, y1, z1,
				imageXIndex, imageYIndex, contrib)) ; 

			if (y2 < gridSide)
			{
				contrib = GetContributions((x1*2.0f)/(gridSide-1)-1.0f, 
								(y2*2.0f)/(gridSide-1)-1.0f,
								(z1*2.0f)/(gridSide-1)-1.0f, 
								pStart, pNext, 2.0f/(gridSide-1),
								rayLength) ; 
				vectGridContributions.push_back(TrilinearGridContribution(x1, y2, z1,
					imageXIndex, imageYIndex, contrib)) ; 
			}
		}
		else
		{
			// all are the same ? Cannot be. 
			throw "In ComputePointContribution with inprobable values" ; 
		}

		xIndexStart = xIndexNext ; yIndexStart = yIndexNext ; zIndexStart = zIndexNext ; 
	}
}

