//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include "TrilinearGridModel.h"
#include "RotationMatrix.h"
#include "ProjectionData.h" 
#include <iostream> 
#include <algorithm> 
#include <sstream>
#include <errno.h> 
#include <float.h>
#include <time.h>


const float LG_SQRT_TWO_PI = 0.5f * log(8.0f*(float)atan(1.0) ) ; 


TrilinearGridModel::TrilinearGridModel(
		const std::vector<RotationMatrix> &vectRotMatrices,
		unsigned char gridSide, 
		unsigned char imageSide): mbyteGridSide(gridSide), mbyteImageSide(imageSide),
			mvectRotations(vectRotMatrices)
{
	mblnUseRandomRotations = false ; 
}
void TrilinearGridModel::SetRotations(const std::vector<RotationMatrix> &vectRotMatrices)
{
	mvectRotations.clear() ; 
	mvectRotations.insert(mvectRotations.begin(), vectRotMatrices.begin(), vectRotMatrices.end()); 
}
TrilinearGridModel::TrilinearGridModel(char *fileName)
{
}

TrilinearGridModel::~TrilinearGridModel(void)
{
}


int CopyWithoutContributionDuplicates(std::vector<TrilinearGridContribution> &vectTarget, 
	const std::vector<TrilinearGridContribution> &vectSource)
{
	// for each guy, we believe that it can be at most 8 pixels before, or 8 pixels after 
	// itself. 
	int numContributions = vectSource.size() ; 
	std::vector<int> vectIndices(vectSource.size()) ; 

	//vectTarget.reserve(vectSource.size()) ; 
	for (int index = 0 ; index < numContributions ; index++)
	{
		const TrilinearGridContribution &contrib = vectSource[index] ; 
		int startIndex = std::max<int>(index-8, 0) ; 
		int correctIndex = -1 ; 
		for ( ; startIndex < index ; startIndex++)
		{
			const TrilinearGridContribution &nextContrib = vectSource[startIndex] ; 
			if (nextContrib.mbyteXIndex == contrib.mbyteXIndex &&
				nextContrib.mbyteYIndex == contrib.mbyteYIndex &&
				nextContrib.mbyteZIndex == contrib.mbyteZIndex &&
				nextContrib.mbyteImageXIndex == contrib.mbyteImageXIndex && 
				nextContrib.mbyteImageYIndex == contrib.mbyteImageYIndex)
			{
				correctIndex = vectIndices[startIndex] ; 
				break ; 
			}
		}
		if (correctIndex != -1)
		{
			// this index was already stored. Just put the contribution in there.
			vectTarget[correctIndex].mdblContribution 
				+= vectSource[index].mdblContribution ; 
			vectIndices[index] = correctIndex ; 
		}
		else
		{
			vectIndices[index] = vectTarget.size() ; 
			vectTarget.push_back(contrib) ; 
		}
	}
	return 0 ; 
}

void TrilinearGridModel::GetGridContributions(
		const unsigned int rotMatrixIndex, 
		std::vector<TrilinearGridContribution> &vectGridContributions) 
{
	if (rotMatrixIndex > mvectRotations.size()-1)
		throw "Wanted rotation index greater than number of matrices provided" ; 

	RotationMatrix rotMatrix = mvectRotations[rotMatrixIndex] ; 
	std::vector<Point> vectIntersectPoints ; 
	vectIntersectPoints.reserve(1024) ; 

	vectGridContributions.clear() ; 
	vectGridContributions.reserve(mbyteImageSide*mbyteImageSide*mbyteImageSide*8) ; 

	std::vector<TrilinearGridContribution> vectGridContributionsForPt ; 
	vectGridContributionsForPt.reserve(1024) ; 

	int numCopied = 0 ; 
	for (unsigned char xPix = 0 ; xPix < mbyteImageSide ; xPix++)
	{
		float x = -1.0f + (2.0f*xPix)/(mbyteImageSide-1) ; 
		for (int yPix = 0 ; yPix < mbyteImageSide ; yPix++)
		{
			vectIntersectPoints.clear() ; 
			vectGridContributionsForPt.clear() ; 

			float y = -1.0f + (2.0f*yPix)/(mbyteImageSide-1) ;
			GridIntersection::ComputeGridIntersectionPoints(vectIntersectPoints, 
				rotMatrix, 
				x, y, 
				mbyteGridSide) ; 
			TrilinearGridContribution::ComputePointContributions(xPix, yPix, 
				mbyteGridSide,
				vectIntersectPoints,
				vectGridContributionsForPt, 
				rotMatrix) ; 
			
			//if ((rotMatrixIndex == 0 || rotMatrixIndex == 1)
			//	&& (xPix == mbyteImageSide/2 && xPix == yPix))
			//{
			//	std::cerr<<"RotMatrix="<<rotMatrixIndex<<std::endl ; 
			//	for (int i = 0 ; i < vectGridContributionsForPt.size() ; i++)
			//	{
			//		const TrilinearGridContribution &pt = vectGridContributionsForPt[i] ;
			//		std::cerr<<(int)pt.mbyteImageXIndex<<","<<(int)pt.mbyteImageYIndex<<"," ; 
			//		std::cerr<<(int)pt.mbyteXIndex<<","<<(int)pt.mbyteYIndex<<","<<(int)pt.mbyteZIndex<<","<<pt.mdblContribution<<std::endl ; 
			//	}
			//}
			vectGridContributions.insert(vectGridContributions.end(), vectGridContributionsForPt.begin(),
				vectGridContributionsForPt.end()) ; 
			//numCopied += CopyWithoutContributionDuplicates(vectGridContributions, 
			//	vectGridContributionsForPt) ; 
		}
	}
}
