//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly


#include "TrilinearGridModelProbabilityComputer.h"
#include "RotationMatrix.h"
#include "ProjectionData.h" 
#include <iostream> 
#include <algorithm> 
#include <sstream>
#include <errno.h> 
#include <float.h>
#include <fstream>
#include <iterator>
#include "Helpers.h" 

const double LG_SQRT_TWO_PI = 0.5 * log(8.0*atan(1.0) ) ; 
const double PI = 4*atan(1.0) ; 
const double SQRT_TWO_PI = sqrt(8.0*atan(1.0) ) ; 


TrilinearGridModelProbabilityComputer::TrilinearGridModelProbabilityComputer(
	const std::vector<RotationMatrix> &vectRotMatrices,
	unsigned char gridSide, 
	unsigned char imageSide): 
		mobjModel(vectRotMatrices, gridSide, imageSide), 
		mdblErrorSigma(1.0f), mbyteGridSide(gridSide),
		mbyteImageSide(imageSide), 
		marrCurrentProjections(NULL),
		marrProjection(NULL), 
		marrProjectionErrors(NULL), 
		MAX_DIST_SQ_FROM_CENTER_IMAGE(0.9f), // the top few pixels start interfering in wierd ways, 
											// so we exclude them from consideration
		MAX_DIST_SQ_FROM_CENTER_GRID(0.9f),// the top few pixels start interfering in wierd ways
		mintCallNum(1)
{
	marrImageBetaGradients = 0 ; 
	marrImageBetaGradientsTemp = 0 ; 
	AllocateMemory() ; 
	GenerateRandomModel() ; 
}


void TrilinearGridModelProbabilityComputer::DeAllocateMemory() 
{
	if (marrBetas != 0)
	{
		for (int numX = 0 ; numX < mbyteGridSide ; numX++)
		{
			for (int numY = 0 ; numY < mbyteGridSide ; numY++)
			{
				if (marrBetas[numX][numY] != 0)
				{
					delete [] marrBetas[numX][numY] ;
					marrBetas[numX][numY] = 0 ; 
				}
				if (marrBetasTotalGradients[numX][numY] != 0)
				{
					delete [] marrBetasTotalGradients[numX][numY] ; 
					marrBetasTotalGradients[numX][numY] = 0 ; 
				}
			}
			if (marrBetas[numX] != 0)
			{
				delete [] marrBetas[numX] ; 
				marrBetas[numX] = 0 ; 
			}
			if (marrBetasTotalGradients[numX] != 0)
			{
				delete [] marrBetasTotalGradients[numX] ; 
				marrBetasTotalGradients[numX] = 0 ; 
			}
		}
		delete [] marrBetas ; 
		delete [] marrBetasTotalGradients ; 
		marrBetas = 0 ; 
		marrBetasTotalGradients = 0 ; 
	}
}

void TrilinearGridModelProbabilityComputer::AllocateMemory() 
{
	marrBetas = new double** [mbyteGridSide] ; 
	marrBetasTotalGradients = new double** [mbyteGridSide] ; 
	for (int numX = 0 ; numX < mbyteGridSide ; numX++)
	{
		marrBetas[numX] = new double * [mbyteGridSide] ; 
		marrBetasTotalGradients[numX] = new double * [mbyteGridSide] ; 
		for (int numY = 0 ; numY < mbyteGridSide ; numY++)
		{
			marrBetas[numX][numY] = new double[mbyteGridSide] ; 
			marrBetasTotalGradients[numX][numY] = new double[mbyteGridSide] ; 
		}
	}
}

TrilinearGridModelProbabilityComputer::~TrilinearGridModelProbabilityComputer(void)
{
	DeAllocateMemory() ; 
	DeAllocateMemoryForImageGradients() ; 
}

void TrilinearGridModelProbabilityComputer::PrintCurrentScores()
{
	std::stringstream ss ;
	std::string scoresFileName = mstrAnalysisName + "_" ;
	ss<<std::string("iter")<<mintCallNum<<std::string("_scores.csv") ; 
	std::string temp  ; 
	ss>>temp ; 
	scoresFileName += temp ; 
	PrintModelScores(scoresFileName) ; 
}

void TrilinearGridModelProbabilityComputer::PrintModelScores(std::string fileName)
{
	std::ofstream fout(fileName.c_str()) ; 
	int index = 0 ; 
	int numRotations = mobjModel.GetNumRotations() ; 
	for (int rotNum = 0 ; rotNum < numRotations ; rotNum++)
	{
		for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
		{
			fout<<mvectImageErrorSquareForRotations[index++] ; 
			if (imageNum != mshortNumImagesInData-1)
				fout<<"," ; 
		}
		fout<<std::endl ; 
	}
	fout.close() ; 
}
void TrilinearGridModelProbabilityComputer::PrintBestRotations(std::string fileName)
{
	int numRotations = mobjModel.GetNumRotations() ; 
	std::vector<double> vectBestScores ; 
	std::vector<int> vectBestScoresIndex ; 

	for (int i = 0 ; i < mshortNumImagesInData ; i++)
	{
		vectBestScores.push_back(DBL_MAX) ; 
		vectBestScoresIndex.push_back(-1) ; 
	}

	int index = 0 ; 
	for (int rotNum = 0 ; rotNum < numRotations ; rotNum++)
	{
		for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
		{
			if(mvectImageErrorSquareForRotations[index] < vectBestScores[imageNum])
			{
				vectBestScoresIndex[imageNum] = rotNum ; 
				vectBestScores[imageNum] = mvectImageErrorSquareForRotations[index] ; 
			}
			index++ ; 
		}
	}
	std::ofstream fout(fileName.c_str()) ; 
	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		int rotationIndex = vectBestScoresIndex[imageNum] ; 
		fout<<mobjModel.GetRotationMatrix(imageNum)<<std::endl ; 
	}
	fout.close() ; 
}

void TrilinearGridModelProbabilityComputer::PrintCurrentModelProjections()
{
	std::stringstream ssProjection ;
	std::string projectionFileName = mstrAnalysisName + "_" ;
	ssProjection<<std::string("iter")<<mintCallNum<<std::string("_projections.mrc") ; 
	std::string temp  ; 
	ssProjection>>temp ; 
	projectionFileName += temp ; 

	CryoData::ProjectionData projectionsData ; 
	projectionsData.SetData(mobjModel.GetNumRotations(), 
		mbyteImageSide, (const float ***) marrCurrentProjections) ; 
	projectionsData.WriteMRC(projectionFileName) ; 
}


void TrilinearGridModelProbabilityComputer::PrintCurrentModel()
{
	std::stringstream ss;
	std::string outFile = mstrAnalysisName + "_" ;
	ss<<std::string("iter")<<mintCallNum<<std::string(".mrc") ; 
	std::string iterStr ; 
	ss>>iterStr ; 
	outFile += iterStr ; 
	PrintModel(outFile) ; 
}

void TrilinearGridModelProbabilityComputer::PrintBestProjections(std::string fileName)
{
	int numRotations = mobjModel.GetNumRotations() ; 
	std::vector<double> vectBestScores ; 
	std::vector<int> vectBestScoresIndex ; 


	for (int i = 0 ; i < mshortNumImagesInData ; i++)
	{
		vectBestScores.push_back(DBL_MAX) ; 
		vectBestScoresIndex.push_back(-1) ; 
	}


	int index = 0 ; 
	for (int rotNum = 0 ; rotNum < numRotations ; rotNum++)
	{
		for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
		{
			if(mvectImageErrorSquareForRotations[index] < vectBestScores[imageNum])
			{
				vectBestScoresIndex[imageNum] = rotNum ; 
				vectBestScores[imageNum] = mvectImageErrorSquareForRotations[index] ; 
			}
			index++ ; 
		}
	}

	std::vector<TrilinearGridContribution> vectGridContributions ; 
	FILE* projFile = CryoData::ProjectionData::OpenMRCForWritingProjections(
		fileName, mbyteImageSide, mbyteImageSide, mshortNumImagesInData) ; 


	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		int rotationIndex = vectBestScoresIndex[imageNum] ; 
		mobjModel.GetGridContributions(rotationIndex, vectGridContributions) ; 
		ComputeCurrentProjection(vectGridContributions) ; 
		CryoData::ProjectionData::AddProjectionToOpenFile(projFile, mbyteImageSide, 
			mbyteImageSide, (const double **&)marrProjection) ; 
		vectGridContributions.clear() ; 
	}
	fclose(projFile) ; 
}

void TrilinearGridModelProbabilityComputer::PrintModel(std::string fileName)
{
	CryoData::ProjectionData outData ;
	outData.SetData(mbyteGridSide, mbyteGridSide, (const double ***)marrBetas, true) ;
	outData.WriteMRC(fileName) ; 
}
void TrilinearGridModelProbabilityComputer::GenerateRandomModel()
{
	srand(clock()) ; 

	for (int numX = 0 ; numX < mbyteGridSide ; numX++)
	{
		for (int numY = 0 ; numY < mbyteGridSide ; numY++)
		{
			for (int numZ = 0 ; numZ < mbyteGridSide ; numZ++)
			{
				marrBetas[numX][numY][numZ] = (double) (rand()*1.0)/RAND_MAX ;  
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::GenerateSphericalModel()
{
	for (int numX = 0 ; numX < mbyteGridSide ; numX++)
	{
		float xVal = -1.0f + 2.0f*numX/(mbyteGridSide-1) ; 
		for (int numY = 0 ; numY < mbyteGridSide ; numY++)
		{
			float yVal = -1.0f + 2.0f*numY/(mbyteGridSide-1) ; 
			for (int numZ = 0 ; numZ < mbyteGridSide ; numZ++)
			{
				float zVal = -1.0f + 2.0f*numZ/(mbyteGridSide-1) ; 
				double dist = sqrt(xVal*xVal + yVal*yVal + zVal*zVal) ; 
				if (dist == 0)
					dist = 2.0/(mbyteGridSide-1) ;  
				if (dist > 1.0)
					marrBetas[numX][numY][numZ] = 0 ; 
				else
					marrBetas[numX][numY][numZ] = .5*(mbyteGridSide-1)/dist ;  
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::SetModelBetas(const unsigned char gridSide, 
		const float  ***data, 
		const unsigned char xColumn, 
		const unsigned char yColumn, 
		const unsigned char zColumn)
{
	if (mbyteGridSide != gridSide)
	{
		DeAllocateMemory() ; 
		AllocateMemory() ; 
	}
	unsigned char columnIndices[3] ; 
	for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
	{
		columnIndices[xColumn] = numX ; 
		for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
		{
			columnIndices[yColumn] = numY ; 
			for (unsigned char numZ = 0 ; numZ < mbyteGridSide ; numZ++)
			{
				columnIndices[zColumn] = numZ ; 
				marrBetas[numX][numY][numZ] = data[columnIndices[0]][columnIndices[1]][columnIndices[2]] ;  
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::SetModelBetas(const unsigned char gridSide, 
		const double ***data, 
		const unsigned char xColumn, 
		const unsigned char yColumn, 
		const unsigned char zColumn)
{
	if (mbyteGridSide != gridSide)
	{
		DeAllocateMemory() ; 
		AllocateMemory() ; 
	}
	unsigned char columnIndices[3] ; 
	for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
	{
		columnIndices[xColumn] = numX ; 
		for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
		{
			columnIndices[yColumn] = numY ; 
			for (unsigned char numZ = 0 ; numZ < mbyteGridSide ; numZ++)
			{
				columnIndices[zColumn] = numZ ; 
				marrBetas[numX][numY][numZ] = data[columnIndices[0]][columnIndices[1]][columnIndices[2]] ;  
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::AllocateMemoryForImageGradients(int numImages) 
{
	if (mshortNumImagesInData == numImages && marrImageBetaGradients != 0)
		return ; 
	if (marrImageBetaGradients != 0)
		DeAllocateMemoryForImageGradients() ; 

	mshortNumImagesInData = numImages ; 
	marrImageBetaGradients = new double*** [mshortNumImagesInData] ; 
	marrImageBetaGradientsTemp = new double*** [mshortNumImagesInData] ; 
	for (short imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		marrImageBetaGradients[imageNum] = new double** [mbyteGridSide] ; 
		marrImageBetaGradientsTemp[imageNum] = new double** [mbyteGridSide] ; 
		for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
		{
			marrImageBetaGradients[imageNum][numX] = new double* [mbyteGridSide] ; 
			marrImageBetaGradientsTemp[imageNum][numX] = new double* [mbyteGridSide] ; 
			for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
			{
				marrImageBetaGradients[imageNum][numX][numY] = 
					new double [mbyteGridSide] ; 
				marrImageBetaGradientsTemp[imageNum][numX][numY] = 
					new double [mbyteGridSide] ; 
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::DeAllocateMemoryForImageGradients()
{
	if (marrImageBetaGradients != 0)
	{
		for (short imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
		{
			for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
			{
				for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
				{
					delete [] marrImageBetaGradients[imageNum][numX][numY] ; 
					delete [] marrImageBetaGradientsTemp[imageNum][numX][numY] ; 
				}
				delete [] marrImageBetaGradients[imageNum][numX] ; 
				delete [] marrImageBetaGradientsTemp[imageNum][numX] ; 
			}
			delete [] marrImageBetaGradients[imageNum] ; 
			delete [] marrImageBetaGradientsTemp[imageNum] ; 
		}
		delete [] marrImageBetaGradients ;
		delete [] marrImageBetaGradientsTemp ;
		marrImageBetaGradients = 0 ; 
		marrImageBetaGradientsTemp = 0 ; 
	}

}

void TrilinearGridModelProbabilityComputer::AllocateMemoryForCurrentProjections()
{
	if (marrProjection != NULL)
		return ; 
	DeAllocateMemoryForCurrentProjections() ; 
	marrProjection = new double* [mbyteImageSide] ; 
	marrProjectionErrors = new double* [mbyteImageSide] ; 
	for (unsigned char x = 0 ; x < mbyteImageSide ; x++)
	{
		marrProjection[x] = new double [mbyteImageSide] ; 
		marrProjectionErrors[x] = new double [mbyteImageSide] ; 
	}
}
void TrilinearGridModelProbabilityComputer::DeAllocateMemoryForCurrentProjections()
{
	if (marrProjection != NULL)
	{
		for (unsigned char x = 0 ; x < mbyteImageSide ; x++)
			delete [] marrProjection[x] ; 
		delete [] marrProjection ; 
	}
	marrProjection = NULL ; 
}

void TrilinearGridModelProbabilityComputer::AllocateMemoryForProjectionImages()
{
	if (marrCurrentProjections != NULL)
		return ; 

	marrCurrentProjections = new double ** [mobjModel.GetNumRotations()] ; 
	for (int rotNum = 0 ; rotNum < mobjModel.GetNumRotations() ; rotNum++)
	{
		marrCurrentProjections[rotNum] = new double * [mbyteImageSide] ; 
		for (short xPix = 0 ; xPix < mbyteImageSide ; xPix++)
				marrCurrentProjections[rotNum][xPix] = new double [mbyteImageSide] ; 
	}
}

void TrilinearGridModelProbabilityComputer::DeAllocateMemoryForProjectionImages()
{
	if (marrCurrentProjections == NULL) 
		return ; 
	for (int rotNum = 0 ; rotNum < mobjModel.GetNumRotations() ; rotNum++)
	{
		for (short xPix = 0 ; xPix < mbyteImageSide ; xPix++)
				delete [] marrCurrentProjections[rotNum][xPix] ; 
		delete [] marrCurrentProjections[rotNum] ; 
	}
	delete [] marrCurrentProjections ; 
	marrCurrentProjections = NULL ; 
}


void TrilinearGridModelProbabilityComputer::ClearMemoryForProjectionGradients(
	bool clearTemp, 
	bool clearGrads,
	bool clearTotalGradients)
{
	if (clearGrads)
	{
		for (short imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
			for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
				for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
					for (unsigned char numZ = 0 ; numZ < mbyteGridSide ; numZ++)
						marrImageBetaGradients[imageNum][numX][numY][numZ] = 0 ; 
	}
	if (clearTemp)
	{
		for (short imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
			for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
				for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
					for (unsigned char numZ = 0 ; numZ < mbyteGridSide ; numZ++)
						marrImageBetaGradientsTemp[imageNum][numX][numY][numZ] = 0 ; 
	}
	if (clearTotalGradients)
	{
		for (unsigned char numX = 0 ; numX < mbyteGridSide ; numX++)
			for (unsigned char numY = 0 ; numY < mbyteGridSide ; numY++)
				for (unsigned char numZ = 0 ; numZ < mbyteGridSide ; numZ++)
					marrBetasTotalGradients[numX][numY][numZ] = 0 ; 
	}
}

void TrilinearGridModelProbabilityComputer::SetModelBetas(const double *(&betas))
{
	int index = 0 ; 
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		float xVal = -1.0f + 2.0f*xIndex/(mbyteGridSide-1) ; 
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			float yVal = -1.0f + 2.0f*yIndex/(mbyteGridSide-1) ; 
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				float zVal = -1.0f + 2.0f*zIndex/(mbyteGridSide-1) ; 
				if (xVal*xVal + yVal*yVal + zVal*zVal < MAX_DIST_SQ_FROM_CENTER_GRID)
					marrBetas[xIndex][yIndex][zIndex] = betas[index++] ; 
				else
				{
					marrBetas[xIndex][yIndex][zIndex] = 0 ; 
					index++ ; 
				}
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::CopyBetas(double * (&betas))
{
	int index = 0 ; 
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		float xVal = -1.0f + 2.0f*xIndex/(mbyteGridSide-1) ; 
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			float yVal = -1.0f + 2.0f*yIndex/(mbyteGridSide-1) ; 
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				float zVal = -1.0f + 2.0f*zIndex/(mbyteGridSide-1) ; 
				if (xVal*xVal + yVal*yVal + zVal*zVal < MAX_DIST_SQ_FROM_CENTER_GRID)
					betas[index++] = marrBetas[xIndex][yIndex][zIndex] ; 
				else
				{
					betas[index++] = 0 ; 
				}
			}
		}
	}
}


void TrilinearGridModelProbabilityComputer::ComputeCurrentProjection(
	const std::vector<TrilinearGridContribution> &vectGridContributions)
{
	for (int x = 0 ; x < mbyteImageSide ; x++)
		for (int y = 0 ; y < mbyteImageSide ; y++)
			marrProjection[x][y] = 0.0f ; 

	for (std::vector<TrilinearGridContribution>::const_iterator iter = vectGridContributions.begin() ; 
			iter != vectGridContributions.end(); iter++)
	{
		const TrilinearGridContribution &cont = (*iter) ; 
		float xVal = -1.0f + 2.0f*cont.mbyteXIndex/(mbyteGridSide-1) ; 
		float yVal = -1.0f + 2.0f*cont.mbyteYIndex/(mbyteGridSide-1) ; 
		float zVal = -1.0f + 2.0f*cont.mbyteZIndex/(mbyteGridSide-1) ; 
		float distanceSq = xVal*xVal+yVal*yVal+zVal*zVal ; 

		if (distanceSq >= MAX_DIST_SQ_FROM_CENTER_GRID)
			continue ; 
		float xValImage = -1.0f + 2.0f*cont.mbyteImageXIndex/(mbyteGridSide-1) ; 
		float yValImage = -1.0f + 2.0f*cont.mbyteImageYIndex/(mbyteGridSide-1) ; 
		if (xValImage*xValImage + yValImage*yValImage >= MAX_DIST_SQ_FROM_CENTER_IMAGE)
			continue ; 
		marrProjection[cont.mbyteImageXIndex][cont.mbyteImageYIndex] += 
			marrBetas[cont.mbyteXIndex][cont.mbyteYIndex][cont.mbyteZIndex] * cont.mdblContribution ;
	}
}

void TrilinearGridModelProbabilityComputer::ComputeRotationGradientsWithoutProbabilityProduct(
		const int rotationNum, 
		std::vector<TrilinearGridContribution> &vectGridContributions,
		std::vector<double> &vectImageErrorSquare, 
		const float ***IxyDataPtr)
{
	int imageNumPixels = mbyteImageSide * (int) mbyteImageSide ; 

	ComputeCurrentProjection(vectGridContributions) ; 

	for (int xIndex = 0 ; xIndex < mbyteImageSide ; xIndex++)
	{
		float xVal = -1.0f + 2.0f*xIndex/(mbyteImageSide-1) ; 
		for (int yIndex = 0 ; yIndex < mbyteImageSide ; yIndex++)
		{
			float yVal = -1.0f + 2.0f*yIndex/(mbyteImageSide-1) ; 
			// is the pixel in the radius area ? 
			float distCenterSq = xVal*xVal + yVal*yVal ; 
			if (distCenterSq >= MAX_DIST_SQ_FROM_CENTER_IMAGE)
				continue ;
			for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
			{

				double pixelError = marrProjection[xIndex][yIndex] - IxyDataPtr[imageNum][xIndex][yIndex] ; 
				vectImageErrorSquare[imageNum] += pixelError * pixelError ; 
			}
			if (mblnSaveCurrentProjections)
				marrCurrentProjections[rotationNum][xIndex][yIndex] = marrProjection[xIndex][yIndex] ; 
		}
	}

	for (std::vector<TrilinearGridContribution>::const_iterator iter = vectGridContributions.begin() ; 
			iter != vectGridContributions.end(); iter++)
	{
		TrilinearGridContribution cont = (*iter) ; 
		// is the pixel in the radius area ? 
		float xVal = -1.0f + 2.0f*cont.mbyteXIndex/(mbyteGridSide-1) ; 
		float yVal = -1.0f + 2.0f*cont.mbyteYIndex/(mbyteGridSide-1) ; 
		float zVal = -1.0f + 2.0f*cont.mbyteZIndex/(mbyteGridSide-1) ; 
		float distanceSq = xVal*xVal+yVal*yVal+zVal*zVal ; 

		if (distanceSq >= MAX_DIST_SQ_FROM_CENTER_GRID)
			continue ; 
		float xValImage = -1.0f + 2.0f*cont.mbyteImageXIndex/(mbyteGridSide-1) ; 
		float yValImage = -1.0f + 2.0f*cont.mbyteImageYIndex/(mbyteGridSide-1) ; 
		if (xValImage*xValImage + yValImage*yValImage >= MAX_DIST_SQ_FROM_CENTER_IMAGE)
			continue ; 
		for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
		{
			if (imageNum == 0 && rotationNum > 0 && mblnFixImage1)
			{
				// the first rotation is fixed for image 1. 
				continue ;
			}
			double pixelError = marrProjection[cont.mbyteImageXIndex][cont.mbyteImageYIndex] 
				- IxyDataPtr[imageNum][cont.mbyteImageXIndex][cont.mbyteImageYIndex] ; 
			marrImageBetaGradientsTemp[imageNum][cont.mbyteXIndex][cont.mbyteYIndex][cont.mbyteZIndex] 
				-= pixelError * cont.mdblContribution ; 
		}
	}
}

void TrilinearGridModelProbabilityComputer::ComputeRotationGradientsWithProbabilityProduct(
	const int rotationNum, 
	const std::vector<double> &vectImageErrorSquare, 
	std::vector<double> &vectImageProbabilityProductPart,
	std::vector<double> &vectImageProbabilityExpPart,
	std::vector<double> &vectImageVarianceProductPart)
{
	const double errVariance = mdblErrorSigma * mdblErrorSigma ; 

	// computation is done. Now normalize image gradients by probabilities. 
	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		// the first image has a fixed rotation - the first one. All other rotations have
		// 0 probability mass for this first image. 
		if (imageNum == 0 && rotationNum > 0 && mblnFixImage1)
			continue ;
		if (rotationNum == 0)
		{

			vectImageProbabilityExpPart[imageNum] = vectImageErrorSquare[imageNum] ;
			vectImageProbabilityProductPart[imageNum] = 1 ; 
			vectImageVarianceProductPart[imageNum] = vectImageErrorSquare[imageNum] ;

			for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
			{
				for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
				{
					for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
					{
						marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] 
							= marrImageBetaGradientsTemp[imageNum][xIndex][yIndex][zIndex]; 
					}
				}
			}
		}
		else
		{
			if (vectImageErrorSquare[imageNum] > vectImageProbabilityExpPart[imageNum])
			{
				double errorSqDiff = -vectImageErrorSquare[imageNum]+vectImageProbabilityExpPart[imageNum] ; 
				double normalizer = exp(errorSqDiff/(2*errVariance)) ; 
				vectImageProbabilityProductPart[imageNum] =  
					vectImageProbabilityProductPart[imageNum] + normalizer; 
				vectImageVarianceProductPart[imageNum] =  
					vectImageVarianceProductPart[imageNum] 
					+ vectImageErrorSquare[imageNum] * normalizer; 
				for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
				{
					for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
					{
						for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
						{
							marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] 
								= marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex]
									+ marrImageBetaGradientsTemp[imageNum][xIndex][yIndex][zIndex] * normalizer ; 
						}
					}
				}
			}
			else
 			{
				double errorSqDiff = vectImageErrorSquare[imageNum]-vectImageProbabilityExpPart[imageNum] ; 
				double normalizer = exp(errorSqDiff/(2*errVariance)) ; 

				vectImageProbabilityProductPart[imageNum] = 
					1 + normalizer * vectImageProbabilityProductPart[imageNum] ; 
				vectImageVarianceProductPart[imageNum] = 
					vectImageErrorSquare[imageNum] + normalizer * vectImageVarianceProductPart[imageNum] ; 
				
				vectImageProbabilityExpPart[imageNum] = vectImageErrorSquare[imageNum] ;
				for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
				{
					for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
					{
						for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
						{
							marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] 
								= marrImageBetaGradientsTemp[imageNum][xIndex][yIndex][zIndex] 
								+ marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] * normalizer ; 
						}
					}
				}
				errno = 0 ; 
			}
		}
	}
}

void TrilinearGridModelProbabilityComputer::ComputeRotationGradientsWithProbabilityProductFixedImages(
	const int rotationNum, 
	const std::vector<double> &vectImageErrorSquare, 
	std::vector<double> &vectImageProbabilityProductPart,
	std::vector<double> &vectImageProbabilityExpPart,
	std::vector<double> &vectImageVarianceProductPart)
{
	// ASSUMING THAT rotationNum is the correct rotation for imageNum
	// and all others are wrong. (i.e. this has probability 1, all others
	// have probability 0.
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				marrImageBetaGradients[rotationNum][xIndex][yIndex][zIndex] 
					= marrImageBetaGradientsTemp[rotationNum][xIndex][yIndex][zIndex]; 
			}
		}
	}
	vectImageVarianceProductPart[rotationNum] = vectImageErrorSquare[rotationNum] ; 
	vectImageProbabilityProductPart[rotationNum] = 1 ; 
	vectImageProbabilityExpPart[rotationNum] =  vectImageErrorSquare[rotationNum] ; 
}

// function calculates prior gradients. If blnApply is true, then 
// the gradients are added to currently computed gradients. 
double TrilinearGridModelProbabilityComputer::CalculatePriorGradients(bool blnApply,
																 double eps)
{
	if (mdblPriorLambda == 0)
		return 0 ; 
	double priorLogProb = 0 ; 
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				int numBoundaries = 0 ;
				double priorDiff = 0 ; 
				if (xIndex > 0)
				{
					double diff = (marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex-1][yIndex][zIndex]) ;
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (yIndex > 0)
				{
					double diff = 
						(marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex][yIndex-1][zIndex]) ; 
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (zIndex > 0)
				{
					double diff = 
						(marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex][yIndex][zIndex-1]) ; 
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (xIndex < mbyteGridSide-1)
				{
					double diff = 
						(marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex+1][yIndex][zIndex]) ; 
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (yIndex < mbyteGridSide-1)
				{
					double diff = 
						(marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex][yIndex+1][zIndex]) ; 
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (zIndex < mbyteGridSide-1)
				{
					double diff = 
						(marrBetas[xIndex][yIndex][zIndex] 
						- marrBetas[xIndex][yIndex][zIndex+1]) ; 
					priorDiff += diff; 
					priorLogProb -= mdblPriorLambda * diff*diff ;
					numBoundaries++; 
				}
				if (blnApply)
				{
					marrBetasTotalGradients[xIndex][yIndex][zIndex] -= 
						eps * 4 * priorDiff * mdblPriorLambda ; 
				}
			}
		}
	}
	return priorLogProb ;
}

double TrilinearGridModelProbabilityComputer::AddPriorRegularization(bool blnApply, double eps)
{
	if (mdblRegularizer == 0)
	{
		return 0 ; 
	}
	double logReg = log(mdblRegularizer); 
	double logProb = 0 ; 
	short PRIOR = 1 ; // 1 for gaussian, 2 for exponential mixed with uniform. 
//	float minVal = 3*mdblRegularizer ; 
	float minVal = 0 ; 
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		float xCoord = -1.0f + 2.0f*xIndex/(mbyteGridSide-1) ;
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			float yCoord = -1.0f + 2.0f*yIndex/(mbyteGridSide-1) ;
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				float zCoord = -1.0f + 2.0f*zIndex/(mbyteGridSide-1) ;
				if(xCoord * xCoord + yCoord * yCoord + zCoord * zCoord >= MAX_DIST_SQ_FROM_CENTER_GRID)
					continue ; 
				// L2 regularizer
				if (PRIOR == 1)
				{
					// only when no negatives in data. 
					double scale = marrBetas[xIndex][yIndex][zIndex]/mdblRegularizer ; 
					logProb += (- LG_SQRT_TWO_PI - logReg - 0.5*scale*scale) ;
					if (blnApply)
					{
						marrBetasTotalGradients[xIndex][yIndex][zIndex] 
							-= eps * scale/mdblRegularizer ;
					}
				}
				else if (PRIOR == 3)
				{
					// only when no negatives in data. 
					double ZERO_PROB = .5 ; 
					double RANGE = 2 ;
					double scale = marrBetas[xIndex][yIndex][zIndex]/mdblRegularizer ; 
					double priorProb = 0.0 ; 
					if (marrBetas[xIndex][yIndex][zIndex] > minVal)
					{
						priorProb = ZERO_PROB * exp(-0.5*scale*scale)/(mdblRegularizer*SQRT_TWO_PI) + (1-ZERO_PROB)/RANGE ; 
						logProb += log(priorProb) ; 
						if (blnApply)
						{
							marrBetasTotalGradients[xIndex][yIndex][zIndex] 
								-= eps * (1-(1-ZERO_PROB)/(RANGE*priorProb)) * scale/mdblRegularizer ;
						}
					}
					else
					{
						logProb += (log(ZERO_PROB) - LG_SQRT_TWO_PI - logReg - 0.5*scale*scale) ;
						if (blnApply)
						{
							marrBetasTotalGradients[xIndex][yIndex][zIndex] 
								-= eps * scale/mdblRegularizer ;
						}
					}
				}
				else if (PRIOR == 2)
				{
					// only when no negatives in data. 
					double ZERO_PROB = .5 ; 
					double RANGE = 1.0 ;
					// regularizer = 1/lambda, i.e. the rate. 
					double priorProb = 0.0 ; 
					if (marrBetas[xIndex][yIndex][zIndex] > minVal)
					{
						priorProb = ZERO_PROB * exp(-marrBetas[xIndex][yIndex][zIndex]/mdblRegularizer)/mdblRegularizer + (1-ZERO_PROB)/RANGE ; 
					}
					else
					{
						priorProb = ZERO_PROB * exp(-marrBetas[xIndex][yIndex][zIndex]/mdblRegularizer)/mdblRegularizer ; 
					}
					logProb += log(priorProb) ; 
					if (blnApply)
					{
						marrBetasTotalGradients[xIndex][yIndex][zIndex] 
							-= eps*ZERO_PROB*exp(-marrBetas[xIndex][yIndex][zIndex]/mdblRegularizer-2*logReg)/priorProb ; 
					}
				}
			}
		}
	}
	return logProb ; 
}

double TrilinearGridModelProbabilityComputer::TrainModelForOneIteration(
		const CryoData::ProjectionData &projData)
{

	std::vector<double> vectImageErrorSquare(projData.GetNumProjections()) ; 

	int imageSide = projData.GetNumX() ; 
	int numRotations = mobjModel.GetNumRotations() ; 
	const float ***IxyDataPtr = projData.GetDataPtr() ; 
	int errNum = 0 ; 
	const int numPlaces = (int)(log10((float)numRotations)+1) ; 
	ClearMemoryForProjectionGradients(true, true, true) ; 
	mvectImageErrorSquareForRotations.clear() ;

	for (int i = 0 ; i < projData.GetNumProjections() ; i++)
	{
		mvectImageProbabilityProductPart[i] = 0 ; 
		mvectImageVarianceProductPart[i] = 0 ; 
		mvectImageProbabilityExpPart[i] = 0 ; 
	}

	std::cerr<<"Rotation #: "<<std::flush ; 
	std::cerr.width(numPlaces+1) ; 

	std::vector<TrilinearGridContribution> vectGridContributions ; 
	for (int rotationNum = 0 ; rotationNum < numRotations ; rotationNum++)
	{
		if (rotationNum % 5 == 0)
		{
			if (rotationNum != 0 )
			{
				std::cerr.width(0) ; 
				for (int i = 0 ; i <= numPlaces ; i++)
					std::cerr<<"\b" ; 
				std::cerr.width(numPlaces+1) ; 
			}
			std::cerr<<rotationNum<<std::flush ; 
		}
		for (int i = 0 ; i < mshortNumImagesInData ; i++)
			vectImageErrorSquare[i] = 0 ; 

		mobjModel.GetGridContributions(rotationNum, vectGridContributions) ; 

		ComputeRotationGradientsWithoutProbabilityProduct(
			rotationNum, 
			vectGridContributions,
			vectImageErrorSquare, 
			IxyDataPtr) ; 

		if (mblnFixRotations)
		{
			ComputeRotationGradientsWithProbabilityProductFixedImages(
				rotationNum,
				vectImageErrorSquare, 
				mvectImageProbabilityProductPart,
				mvectImageProbabilityExpPart,
				mvectImageVarianceProductPart) ; 
		}
		else
		{
			ComputeRotationGradientsWithProbabilityProduct(
				rotationNum,
				vectImageErrorSquare, 
				mvectImageProbabilityProductPart,
				mvectImageProbabilityExpPart,
				mvectImageVarianceProductPart) ; 
		}

		ClearMemoryForProjectionGradients(true, false, false) ; 

		mvectImageErrorSquareForRotations.insert(mvectImageErrorSquareForRotations.end(), 
			vectImageErrorSquare.begin(), vectImageErrorSquare.end()); 
	}

	// lets recompute error sigma as well. 

	double logLikelihood = 0 ; 
	double computedErrorSigma = sqrt(CalculateNextVariance()) ; 
	double numPixels = PI*mbyteImageSide * mbyteImageSide /4.0 ; 

	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		//logLikelihood = logLikelihood  
		//	- mvectImageProbabilityExpPart[imageNum]/(2*computedErrorSigma*computedErrorSigma)
		//	+ log(mvectImageProbabilityProductPart[imageNum]) 
		//	- imageSide * imageSide * log(computedErrorSigma); 
		//std::cerr<<"ImageNum = "<<imageNum<<" Exp Part = "<<mvectImageProbabilityExpPart[imageNum]<<" Product PArt = "<<mvectImageProbabilityProductPart[imageNum]<<std::endl ;
		logLikelihood = logLikelihood  
			- mvectImageProbabilityExpPart[imageNum]/(2.0*mdblErrorSigma*mdblErrorSigma)
			+ log(mvectImageProbabilityProductPart[imageNum]) 
			- numPixels * log(mdblErrorSigma*1.0); 
	}

	// add the prior. 
	mdblCurrentLogPrior = CalculatePriorGradients(false, 0) + AddPriorRegularization(false, 0) ;
	logLikelihood += mdblCurrentLogPrior  ;

	return logLikelihood ; 
}

double TrilinearGridModelProbabilityComputer::CalculateNextVariance() 
{
	double nextVariance = 0 ; 
	double numPixels = PI * mbyteImageSide * mbyteImageSide / 4 ; 

	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		//std::cerr<<"Image "<<imageNum<<" Variance = "<<mvectImageVarianceProductPart[imageNum]<<" "<<mvectImageProbabilityProductPart[imageNum]<<std::endl; 
		nextVariance += (mvectImageVarianceProductPart[imageNum]/mvectImageProbabilityProductPart[imageNum])/(numPixels* (double)mshortNumImagesInData) ;
	}
	return nextVariance ; 
}

double TrilinearGridModelProbabilityComputer::ComputeLogLikelihood(
		const double *betas,
		double *gradients,
		CryoData::ProjectionData &projData, 
		double regularizer, 
		double priorLambda, 
		bool blnFixImage1,
		bool blnFixRotations, 
		bool blnSaveCurrentProjections,
		const float noiseSD)
{
	std::cerr<<"Function Call #:"<<mintCallNum<<" Noise SD = "<<noiseSD<<" "<<std::flush ; 
	mblnFixImage1 = blnFixImage1 ; 
	mblnFixRotations = blnFixRotations ; 
	mdblPriorLambda = priorLambda ; 
	mdblErrorSigma = noiseSD ; 
	mdblRegularizer = regularizer ; 
	mblnSaveCurrentProjections = blnSaveCurrentProjections ; 
	AllocateMemoryForImageGradients(projData.GetNumProjections()) ; 
	AllocateMemoryForCurrentProjections() ; 
	if (mblnSaveCurrentProjections)
		AllocateMemoryForProjectionImages() ; 

	SetModelBetas((const double *&)betas);
	mvectImageProbabilityProductPart.resize(projData.GetNumProjections()) ; 
	mvectImageVarianceProductPart.resize(projData.GetNumProjections()) ;
	mvectImageProbabilityExpPart.resize(projData.GetNumProjections()) ;

	double ll = TrainModelForOneIteration(projData) ; 
	double normalizer = 1.0/(noiseSD*noiseSD) ; 
	for (int imageNum = 0 ; imageNum < mshortNumImagesInData ; imageNum++)
	{
		for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
		{
			for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
			{
				for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
				{
					marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] /= mvectImageProbabilityProductPart[imageNum] ;
					marrBetasTotalGradients[xIndex][yIndex][zIndex] += marrImageBetaGradients[imageNum][xIndex][yIndex][zIndex] * normalizer ; 
				}
			}
		}
	}
	CalculatePriorGradients(true, 1) ; 
	if (mdblRegularizer != 0)
		AddPriorRegularization(true, 1) ; 

	mdblErrorSigma = sqrt(CalculateNextVariance()) ; 

	int index = 0 ; 
	for (short xIndex = 0 ; xIndex < mbyteGridSide ; xIndex++)
	{
		float xVal = -1.0f + 2.0f*xIndex/(mbyteGridSide-1) ; 
		for (short yIndex = 0 ; yIndex < mbyteGridSide ; yIndex++)
		{
			float yVal = -1.0f + 2.0f*yIndex/(mbyteGridSide-1) ; 
			for (short zIndex = 0 ; zIndex < mbyteGridSide ; zIndex++)
			{
				float zVal = -1.0f + 2.0f*zIndex/(mbyteGridSide-1) ; 
				if (xVal*xVal + yVal*yVal + zVal*zVal < MAX_DIST_SQ_FROM_CENTER_GRID)
					gradients[index++] = -marrBetasTotalGradients[xIndex][yIndex][zIndex] ; 
				else
					gradients[index++] = 0 ; 
			}
		}
	}

	if (mblnSaveCurrentProjections)
		PrintCurrentModelProjections() ; 

	// write out mobjProbabilityModel. 
	//PrintCurrentModel() ; 
	//PrintCurrentScores() ; 
	std::cerr<<" Log likelihood = "<<ll<<" Noise = "<<mdblErrorSigma ; 
	mintCallNum++ ; 
	return ll ; 
}


