//	This file is part of CryoBayes.
//  CryoBayes is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  CryoBayes is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// CryoBayes: Copyright 2009,2010 Navdeep Jaitly
#include <iostream> 
#include <vector>
#include <fstream> 
#include <sstream> 
#include <string> 
#include "LBFGSFitTrilinearModel.h" 
#include "ProjectionData.h"
#include "RotationMatrix.h" 
#include <time.h>
#include <cstdlib>

void Usage(int argc, char **argv)
{
	std::cerr<<"Usage:"<<argv[0]<<" -i [mrc file] -o [analysis name]"<<std::endl ; 
	std::cerr<<"\t-i: Specify the name of the input mrc file."<<std::endl ; 
	std::cerr<<"\t-o: Specify the name of the output analyses (no .mrc extension)."<<std::endl ; 
	std::cerr<<"\t-r: Specify the rotation file to use."<<std::endl ; 
	std::cerr<<"\t-p: Specify the PSF file to use."<<std::endl ; 
	std::cerr<<"\t-s: Specify the mrc file with the starting mobjProbabilityModel (Optional)."<<std::endl ; 
	std::cerr<<"\t -eps: Specify epsilon improvement in log likelihood for continued iterations (Optional. Default=.1)."<<std::endl ; 
	std::cerr<<"\t -nGeo: Specify number of subdivisions for Geodesic points. Default=3)."<<std::endl ; 
	std::cerr<<"\t -psisDeg: Specifies the granularity (in degrees) of the rotations around the z-axis. Default=6)."<<std::endl ; 
	std::cerr<<"\t -nIter: Number of iterations for each annealing run. (Optional. Default = 10)"<<std::endl ; 
	std::cerr<<"\t -reg: Regularizer penalty value. (Optional. Default = 1e-6. Suggested 1e-3 when no psf correction used)"<<std::endl ; 
	std::cerr<<"\t -free: Do not fix image 1 to some position. (Optional. Default = fix 1 image)"<<std::endl ; 
	std::cerr<<"\t -pLamda: Prior lambda for spatial prior (Default = 0.1)"<<std::endl ; 
	std::cerr<<"\t -fixRot: Fix rotations to those provided in rotations file (Only applies when rotation file is provided as argument.)"<<std::endl ; 
	std::cerr<<"\t -sProj: Save Projections. "<<std::endl ; 
	std::cerr<<"\t -minSD: Minimum standard deviation during annealiing (Default = 0.1). "<<std::endl ; 
	std::cerr<<"\t -sSD: Starting standard deviation during annealiing (Default = 20.0). "<<std::endl ; 
	std::exit(1) ;
}

void ParseProgramArguments(int argc, char **argv, std::string &mrcFile, 
						   std::string &analysisName, 
						   std::string &rotFile, std::string &startMRCFile, 
						   float &epsilon, 
						   int &numIterations, 
						   float &regularizer, 
						   bool &blnFixFirstImage,
						   int &numGeodesicPts, 
						   float &psiDeg,
						   float &priorLambda, 
						   bool &blnFixRotations, 
						   bool &blnSaveProjections,
						   float &minSDForAnnealing,
						   float &startSDForAnnealing)
{
	int argNum = 1; 
	const std::string mrcArgName("-i") ; 
	const std::string analysisArgName("-o") ; 
	const std::string rotationArgName("-r") ; 
	const std::string startMRCArgName("-s") ; 
	const std::string epsArgName("-eps") ; 
	const std::string numGeodesicPtsArgName("-nGeo") ; 
	const std::string psiDegArgName("-psisDeg") ; 
	const std::string numIterationsArgName("-nIter") ; 
	const std::string regArgName("-reg") ; 
	const std::string fixArgName("-free") ; 
	const std::string priorLambdaArgName("-pLambda") ; 
	const std::string fixRotArgName("-fixRot") ; 
	const std::string blnSaveProjectionArgName("-sProj") ; 
	const std::string minSDArgName("-minSD") ; 
	const std::string startSDArgName("-sSD") ; 

	try 
	{
		while (argNum < argc)
		{
			std::string arg(argv[argNum]) ; 
			if (arg.compare(mrcArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"MRC File not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				mrcFile.assign(argv[argNum]) ; 
				++argNum ;
			}
			else if (arg.compare(analysisArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Analysis name not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				analysisName.assign(argv[argNum]) ; 
				++argNum ;
			}
			else if (arg.compare(rotationArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Rotation File not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				rotFile.assign(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(startMRCArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Starting File not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				startMRCFile.assign(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(epsArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Epsilon not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				epsilon = (float) atof(argv[argNum]); 
				++argNum ;

			}
			else if (arg.compare(minSDArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Minimum annealing standard deviation not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				minSDForAnnealing = (float) atof(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(startSDArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Starting annealing standard deviation not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				startSDForAnnealing = (float) atof(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(regArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Regularizer value not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				regularizer = (float) atof(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(numGeodesicPtsArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Number of geodesic subdivisions not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				numGeodesicPts = atoi(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(psiDegArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Psi degrees not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				psiDeg = (float) atof(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(priorLambdaArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Prior Lambda not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				priorLambda = (float) atof(argv[argNum]); 
				++argNum ;
			}

			else if (arg.compare(numIterationsArgName) == 0)
			{
				++argNum ; 
				if (argNum == argc)
				{
					std::cerr<<"Number of iterations not specified"<<std::endl ; 
					Usage(argc, argv) ; 
				}
				numIterations = atoi(argv[argNum]); 
				++argNum ;
			}
			else if (arg.compare(blnSaveProjectionArgName) == 0)
			{
				++argNum ; 
				blnSaveProjections = true ; 
			}
			else if (arg.compare(fixArgName) == 0)
			{
				++argNum ; 
				blnFixFirstImage = false ; 
			}
			else if (arg.compare(fixRotArgName) == 0)
			{
				++argNum ; 
				blnFixRotations = true ; 
			}
			else 
			{
				std::cerr<<"Incorrect argument: "<<argv[argNum]<<std::endl ; 
				Usage(argc, argv) ; 
			}
		}
	}
	catch (char *err)
	{
		std::cerr<<"Error parsing arguments: "<<err<<std::endl ; 
		exit(1) ; 
	}
	if (mrcFile.length() == 0)
	{
		std::cerr<<"Data file not specified."<<std::endl ; 
		Usage(argc, argv) ; 
	}
	if (analysisName.length() == 0)
	{
		std::cerr<<"Analysis name not specified."<<std::endl ; 
		Usage(argc, argv) ; 
	}
}


int main(int argc, char **argv)
{
	srand(clock()) ; 
	std::string mrcFile("") ; 
	std::string rotFile("") ; 
	std::string analysisName("") ; 
	std::string startMRCFile("") ;

	bool blnFixFirstImage = true;  
	int numIterations = 10 ; 
	float epsilon=0.1f ;
	float regularizer = 1e-6f;
	float psiDeg = 6.0f ; 
	int numGeodesicPts = 3 ; 
	float priorLambda = 0.0001f ; 
	bool blnFixRotations = false ;
	bool blnSaveProjections = false ; 
	float minSDForAnnealing = 0.1f ;
	float startSDForAnnealing = 20.0f ;

	ParseProgramArguments(argc, argv, mrcFile, analysisName, rotFile, startMRCFile,
		epsilon, numIterations, regularizer, 
		blnFixFirstImage, numGeodesicPts, psiDeg, priorLambda,
		blnFixRotations, blnSaveProjections, 
		minSDForAnnealing, startSDForAnnealing) ;

	std::cerr<<"Input file: "<<mrcFile<<std::endl ; 
	std::cerr<<"Analysis Name: "<<analysisName<<std::endl ; 
	std::cerr<<"Rotation File: "<<rotFile<<std::endl ;
	std::cerr<<" Starting Model = "<<startMRCFile<<std::endl ; 
	std::cerr<<" Regularizer = "<<regularizer<<std::endl ; 
	std::cerr<<" Fix First Image ? = "<<blnFixFirstImage<<std::endl ; 
	std::cerr<<" Number of iterations = "<<numIterations<<std::endl ; 
	std::cerr<<" Prior Lambda = "<<priorLambda<<std::endl ; 
	std::cerr<<" Fix rotations = "<<blnFixRotations<<std::endl ; 
	std::cerr<<" Save Projections = "<<blnSaveProjections<<std::endl ; 
	std::cerr<<" Starting Annealing Standard deviations = "<<startSDForAnnealing<<std::endl ; 
	std::cerr<<" Minimum Annealing Standard deviations = "<<minSDForAnnealing<<std::endl ; 
	std::cerr<<" Log likelihood improvement Epsilon = "<<epsilon<<std::endl ; 
	//CompareDataToProjections(startMRCFile, mrcFile, rotFile, numGeodesicPts, psiDeg) ; 
	//CreateProjections(std::string("resampled_model_9partsr_zeros.mrc"), 9) ; 
	//CreateProjections(std::string("resampled_model_9partsr_zeros.mrc"), 
	//	std::string("projections_resampled_model_9partsr_zeros_9.mrc"), 
	//	std::string("atpRotations.csv")) ; 

	LBFGSFitTrilinearModel modelFitter;
	modelFitter.FitModel(mrcFile, analysisName, rotFile, 
		startMRCFile,numIterations, regularizer,
		blnFixFirstImage, numGeodesicPts, psiDeg, 
		priorLambda,blnFixRotations, blnSaveProjections,
		startSDForAnnealing, minSDForAnnealing, 
		epsilon) ; 
	return 0 ; 
}
